# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
from copy import copy
import functools
from uuid import uuid4
from toxicbuild.core.exceptions import ExecCmdError
from toxicbuild.core.utils import (exec_cmd, LoggerMixin, datetime2string,
                                   now, string2datetime, localtime2utc)


class Builder(LoggerMixin):

    """ A builder executes build steps. Builders are configured in
    the toxicbuild.conf file
    """

    def __init__(self, manager, name, workdir, platorm='linux-generic',
                 remove_env=True, **envvars):
        """:param manager: instance of :class:`toxicbuild.slave.BuildManager`.
        :param name: name for this builder.
        :param workdir: directory where the steps will be executed.
        :param: platform: When the builder execute is builds.
        :param remove_env: Indicates if the build environment should be
          removed when the build is done.
        :param envvars: Environment variables to be used on the steps.
        """
        self.manager = manager
        self.name = name
        self.workdir = workdir
        self.steps = []
        self.plugins = []
        self.platform = platorm
        self.remove_env = remove_env
        self.envvars = envvars

    async def __aenter__(self):
        await self._copy_workdir()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.remove_env:
            await self._remove_tmp_dir()

    def _run_in_build_env(self):
        # This is basicaly useless. It is here
        # just because async with self._run_in_build_env() looks
        # better than async with self.
        return self

    async def build(self):
        async with self._run_in_build_env():
            build_info = await self._do_build()
        return build_info

    @asyncio.coroutine
    def _do_build(self):
        build_status = None
        build_info = {'steps': [], 'status': 'running',
                      'started': datetime2string(now()),
                      'finished': None, 'info_type': 'build_info'}

        yield from self.manager.send_info(build_info)

        for index, step in enumerate(self.steps):
            msg = 'Executing %s' % step.command
            self.log(msg, level='debug')
            local_now = localtime2utc(now())
            step_info = {'status': 'running', 'cmd': step.command,
                         'name': step.name,
                         'started': datetime2string(local_now),
                         'finished': None, 'index': index, 'output': '',
                         'info_type': 'step_info', 'uuid': str(uuid4()),
                         'total_time': None}

            yield from self.manager.send_info(step_info)

            envvars = self._get_env_vars()

            out_fn = functools.partial(self._send_step_output_info, step_info)

            step_exec_output = yield from step.execute(cwd=self.workdir,
                                                       out_fn=out_fn,
                                                       **envvars)
            step_info.update(step_exec_output)

            status = step_info['status']
            msg = 'Finished {} with status {}'.format(step.command, status)
            self.log(msg, level='debug')

            finished = localtime2utc(now())
            step_info.update({'finished': datetime2string(finished)})
            step_info['total_time'] = (
                finished - string2datetime(step_info['started'])).seconds

            yield from self.manager.send_info(step_info)

            # here is: if build_status is something other than None
            # or success (ie failed) we don't change it anymore, the build
            # is failed anyway.
            if build_status is None or build_status == 'success':
                build_status = status

            build_info['steps'].append(step_info)

            if status == 'fail' and step.stop_on_fail:
                break

        build_info['status'] = build_status
        build_info['total_steps'] = len(self.steps)
        finished = now()
        build_info['finished'] = datetime2string(finished)

        return build_info

    @asyncio.coroutine
    def _send_step_output_info(self, step_info, line_index, line):
        msg = {'info_type': 'step_output_info',
               'uuid': step_info['uuid'], 'output_index': line_index,
               'output': line.strip('\n')}

        yield from self.manager.send_info(msg)

    def _get_env_vars(self):
        envvars = copy(self.envvars)
        for plugin in self.plugins:
            envvars.update(plugin.get_env_vars())
        return envvars

    def _get_tmp_dir(self):
        return '{}-{}'.format(self.workdir, self.name)

    @asyncio.coroutine
    def _copy_workdir(self):
        """Copy a workdir to a temp dir to run the tests"""

        tmp_dir = self._get_tmp_dir()
        self.log('Coping workdir to {}'.format(tmp_dir), level='debug')

        mkdir_cmd = 'mkdir -p {}'.format(tmp_dir)
        yield from exec_cmd(mkdir_cmd, cwd='.')
        cp_cmd = 'cp -R {}/* {}'.format(self.workdir, tmp_dir)

        self.log('Executing {}'.format(cp_cmd), level='debug')
        yield from exec_cmd(cp_cmd, cwd='.')

    @asyncio.coroutine
    def _remove_tmp_dir(self):
        """Removes the temporary dir"""

        self.log('Removing tmp-dir', level='debug')
        rm_cmd = 'rm -rf {}'.format(self._get_tmp_dir())
        self.log('Executing {}'.format(rm_cmd), level='debug')
        yield from exec_cmd(rm_cmd, cwd='.')


class BuildStep:

    def __init__(self, name, command, warning_on_fail=False, timeout=3600,
                 stop_on_fail=False):
        """:param name: name for the command.
        :param cmd: a string the be executed in a shell.
        :param warning_on_fail: Indicates if should have warning status if
          the command fails.
        :param timeout: How long we wait for the command to complete.
        :param stop_on_fail: If True and the step fails the build will stop.
        """
        self.name = name
        self.command = command
        self.warning_on_fail = warning_on_fail
        self.timeout = timeout
        self.stop_on_fail = stop_on_fail

    def __eq__(self, other):
        if not hasattr(other, 'command'):
            return False

        return self.command == other.command

    @asyncio.coroutine
    def execute(self, cwd,  out_fn=None, **envvars):
        """Executes the step command.
        :param cwd: Directory where the command will be executed.
        :param out_fn: Function used to handle each line of the
          command output.
        :param envvars: Environment variables to be used on execution.
        """

        step_status = {}
        try:
            output = yield from exec_cmd(self.command, cwd=cwd,
                                         timeout=self.timeout,
                                         out_fn=out_fn, **envvars)
            status = 'success'
        except ExecCmdError as e:
            output = e.args[0]
            if self.warning_on_fail:
                status = 'warning'
            else:
                status = 'fail'
        except asyncio.TimeoutError:
            status = 'exception'
            output = '{} has timed out in {} seconds'.format(self.command,
                                                             self.timeout)

        step_status['status'] = status
        step_status['output'] = output

        return step_status



# -*- coding: utf-8 -*-

import asyncio
from toxicbuild.core.utils import log
from toxicbuild.slave.protocols import BuildServerProtocol


class BuildServer:

    def __init__(self, addr='0.0.0.0', port=7777):
        self.protocol = BuildServerProtocol
        self.loop = asyncio.get_event_loop()
        self.addr = addr
        self.port = port
        coro = self.loop.create_server(self.get_protocol_instance, addr,  port)
        self.server = self.loop.run_until_complete(coro)

    def get_protocol_instance(self):
        return self.protocol(self.loop)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.close()
        self.loop.run_until_complete(self.server.wait_closed())
        self.loop.close()

    def start(self):
        self.loop.run_forever()


def run_server(addr='0.0.0.0', port=7777):
    log('Serving at {}'.format(port))
    with BuildServer(addr, port) as server:
        server.start()



# -*- coding: utf-8 -*-

# Copyright 2015, 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

# Messy! Have fun trying to debug it. I had. :P
# It's just for fun, if you want to implement another ui to toxicubuild
# you should take a look at toxicbuild.ui.models instead of interacting
# directly with master.hole as is done in ToxicCliAction - or you could
# use it directly, that's your code!

import asyncio
from collections import deque
import os
import random
import re
import urwid
from toxicbuild.core.exceptions import ToxicClientException
from toxicbuild.ui import inutils
from toxicbuild.ui.client import get_hole_client
from toxicbuild.ui.models import User


class ToxicShellError(Exception):
    pass


def parse_cmdline(cmdline):
    """ Parses a command line from toxicshell.
    Returns action, args, kwargs for the command line.

    :param cmdline: Input from the user as a string."""

    # getting rid of the spaces around equal signs (kwargs)
    cmdline = re.sub(re.compile('(\s+|)=(\s+|)'), '=', cmdline)

    # removing quotes
    cmdline = re.sub(re.compile('(\'|")'), '', cmdline)

    action, rest = cmdline.split()[0], cmdline.split()[1:]
    cmdargs, cmdkwargs = [], {}

    for part in rest:
        if '=' in part:
            k, v = part.split('=')
            cmdkwargs.update({k: v})
        else:
            if part.isdigit():
                part = int(part)
            cmdargs.append(part)

    return action, cmdargs, cmdkwargs


def validate_input(command_params, user_args, user_kwargs):
    """ Validates args and kwargs from the user input against
    the args and kwargs of a known command.

    :param command_params: Known parameters for the command.
    :param user_args: Argument list submitted by the user.
    :param user_kwargs: Keyword arguments submitted by the user.
    """

    command_args = [arg['name'] for arg in command_params if arg['required']]
    command_kwargs = {arg['name']: arg['default'] for arg in command_params
                      if arg['name'] not in command_args}

    # validate this can be tricky, so lets python do the job.
    # and how I could not figure out how to create a function based on a
    # function signature in a proper way I'm using this ugly eval thing.

    args = ', '.join(command_args)
    kwargs = ', '.join(['{}={}'.format(k, v)
                        for k, v in command_kwargs.items()])

    sig = ', '.join([args, kwargs]) if (args or kwargs) else ' '
    sig = sig.strip().strip(',')

    f = eval('lambda {}: None'.format(sig))

    try:
        f(*user_args, **user_kwargs)
    except Exception as e:
        raise ToxicShellError(str(e))


def get_kwargs(command_params, args):
    """ Return kwargs for ``args`` based on ``command_params``.
    ``command_params`` are the known parameters for a command, those
    sent by list-funcs."""

    kwargs = {}
    for i, arg in enumerate(args):
        try:
            kwargs.update({command_params[i]['name']: arg})
        except IndexError:
            msg = _('Bad args for command')  # noqa f821
            raise ToxicShellError(msg)

    return kwargs


class HistoryEditMixin:

    def __init__(self, *args, **kwargs):
        # super here so it works in cooperation
        super().__init__(*args, **kwargs)
        self.history_file = os.path.join(os.path.expanduser('~'),
                                         '.toxiccli_history')
        self.history_limit = 100

        self.history = self._get_history()
        self._history_position = -1

    def append_to_history(self, item):
        # If you type the same command repeatedly we will not store
        # all these times on history.
        item = item.strip().strip('\n')
        if (not item or (self.history and item == self.history[-1])):
            return

        if len(self.history) < self.history_limit:
            self.history.append(item)
        else:
            self.history[-1] = item

    def save_history(self):
        with open(self.history_file, 'w') as fd:
            fd.write('\n'.join(self.history))

    def get_previous(self):
        prev = self.history[self._history_position]
        self._history_position -= 1
        if abs(self._history_position) > len(self.history):
            self._history_position = -1
        return prev

    def get_next(self):
        self._history_position += 1
        if self._history_position >= len(self.history):
            self._history_position = 0
        n = self.history[self._history_position]
        return n

    def keypress(self, size, key):
        if key == 'enter':
            self._history_position = -1
            edit_text = self.get_edit_text().strip()
            if edit_text:
                self.append_to_history(edit_text)

            return super().keypress(size, key)

        elif key not in ('up', 'down') or not self.history:
            return super().keypress(size, key)

        if key == 'up':
            cmd = self.get_previous()
        else:
            # down
            cmd = self.get_next()

        self.set_edit_text(cmd)

    def _get_history(self):
        if not os.path.exists(self.history_file):
            history = deque()
        else:
            with open(self.history_file, 'r') as fd:
                lines = [l.strip().strip('\n') for l in fd.readlines()]
                history = deque()
                history.extend(lines)

        return history


class HistoryEdit(HistoryEditMixin, urwid.Edit):
    pass


class ToxicCliActions:

    def __init__(self, username_or_email, password, *args,
                 host='localhost', port=6666, token=None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.username_or_email = username_or_email
        self.password = password
        self.host = host
        self.port = port
        self.token = token
        self.user = None
        self._loop = asyncio.get_event_loop()
        self.actions = self.get_actions()

    async def authenticate(self):
        kw = {'username_or_email': self.username_or_email,
              'password': self.password}

        try:
            client = await self.get_client()
            user_dict = await client.user_authenticate(**kw)
            self.user = User(None, user_dict)
        finally:
            client.disconnect()

    @asyncio.coroutine
    def get_client(self):
        """ Returns a client connected to a toxicbuild master"""

        client = yield from get_hole_client(self.user, self.host, self.port,
                                            hole_token=self.token)
        return client

    def get_actions(self):
        """ Asks the server for which actions are available. """

        if not self.user:
            self._loop.run_until_complete(self.authenticate())

        with self._loop.run_until_complete(self.get_client()) as client:
            actions = self._loop.run_until_complete(client.list_funcs())
        return actions

    def get_action_from_command_line(self, cmdline):
        """ Returns `cmdname` and `cmdkwargs`. """

        cmd, cmdargs, cmdkwargs = parse_cmdline(cmdline)

        try:
            sig = self.actions[cmd]
        except KeyError:
            msg = _('Command "{}" does not exist').format(cmd)  # noqa f821
            raise ToxicShellError(msg)

        known_params = sig['parameters']

        cmdkwargs.update(get_kwargs(known_params, cmdargs))
        return cmd, cmdkwargs

    def get_action_help(self, action_name):
        try:
            action_sig = self.actions[action_name]
        except KeyError:
            msg = _('Command "{}" does not exist').format(action_name)  # noqa f821
            raise ToxicShellError(msg)

        action_help = {}
        action_help['short_doc'] = action_sig['doc'].splitlines()[0]

        action_help['doc'] = action_sig['doc']
        action_help['parameters'] = action_sig['parameters']
        return action_help

    @asyncio.coroutine
    def execute_action(self, cmdline):
        """ Execute some action based on the ``cmdline`` inputted by
        the user."""

        action, cmdkwargs = self.get_action_from_command_line(cmdline)

        with (yield from self.get_client()) as client:
            response = yield from client.request2server(action, cmdkwargs)
        return action, response


class ToxicCli(ToxicCliActions, urwid.Filler):

    # urwid, great library!

    def __init__(self, username_or_email, password, host='localhost',
                 port=6666, token=None):

        self.prompt = 'toxicbuild> '
        self.input = HistoryEdit(self.prompt)
        self.messages = urwid.Text('')
        self.main_screen = urwid.Text('')
        self.div = urwid.Divider()
        self.pile = urwid.Pile([self.main_screen, self.div, self.messages,
                                self.div, self.input])
        super().__init__(username_or_email, password, self.pile,
                         valign='bottom',
                         host=host, port=port,
                         token=token)

        self._stop_peek = False
        self._peek_output = []

    def __getattr__(self, attrname):
        if attrname.startswith('_format'):
            action = attrname.replace('_format_', '').replace('_', '-')

            if action in self.actions.keys():
                return self._format_result
        raise AttributeError(attrname)

    def run(self):
        palette = [('bold_blue', 'dark blue,bold', ''),
                   ('bold_yellow', 'yellow,bold', ''),
                   ('bold_white', 'white,bold', ''),
                   ('bold_red', 'dark red,bold', ''),
                   ('error', 'dark red', ''),
                   ('action-name', 'dark blue,bold', ''),
                   ('param', 'white,bold', '')]

        self.show_welcome_screen()
        evl = urwid.AsyncioEventLoop(loop=asyncio.get_event_loop())
        urwid.MainLoop(self, palette, event_loop=evl).run()

    def quit(self):
        self.input.save_history()
        raise urwid.ExitMainLoop()

    def keypress(self, size, key):
        super().keypress(size, key)
        if key != 'enter':
            return

        input_text = self.input.text.split(self.prompt)[1].strip()
        self.input.set_edit_text('')
        self.messages.set_text('')
        if not input_text:
            return
        asyncio.async(self.execute_and_show(input_text))

    @asyncio.coroutine
    def execute_and_show(self, cmdline):
        """ Executes an action requested by the user and show
        its output. """

        # A bunch of ugly stuff, I know!
        if cmdline in ['help', 'h']:
            action, response = 'help', self.get_help_screen()
        elif cmdline.startswith('help'):
            action, response = 'help', self.get_action_help_screen(
                cmdline.split(' ')[1], full=True)
        elif cmdline in ['quit', 'q']:
            # return just for tests
            return self.quit()
        elif cmdline == 'peek':
            return (yield from self.peek())

        elif cmdline == 'stop-peek':
            self.stop_peek()
            self.main_screen.set_text('stopped!')
            return
        elif cmdline == 'welcome-screen':
            self.show_welcome_screen()
            return
        else:
            try:
                action, response = yield from self.execute_action(cmdline)
            except (ToxicShellError, ToxicClientException) as e:
                self.messages.set_text(('error', str(e)))
                return

        meth_name = '_format_' + action.replace('-', '_')
        format_meth = getattr(self, meth_name)
        response = format_meth(response)

        self.main_screen.set_text(response)

    @asyncio.coroutine
    def peek(self):
        """Peeks throught the master's hole."""

        with (yield from self.get_client()) as client:

            response = yield from client.request2server('stream', {})
            self.main_screen.set_text(self._format_result(response))
            while True:
                response = yield from client.get_response()
                if self._stop_peek:
                    client.diconnect()
                    break
                self.main_screen.set_text(
                    self._format_peek(response))  # pragma no cover

        self._stop_peek = False

    def stop_peek(self):
        self._stop_peek = True

    # from here to eternity are the methods that write the screens
    # or methods related to that.
    def show_welcome_screen(self):
        """ Displays the welcome screen for toxiccli.
        """

        welcome = self._get_welcome_text()
        welcome = welcome.replace('{toxicbuild}', '')

        text = [welcome, '\n']

        toxicbuild = random.choice(inutils.LOGOS)
        text += toxicbuild

        text.append('\n')
        text.append(random.choice(inutils.SENTENCES))
        text.append('\n\n')

        help_text = self._get_help_text()

        formated_help = self._format_help_text(help_text)

        text += formated_help

        text.append('\n\n\n')

        self.main_screen.set_text(text)

    def get_action_help_screen(self, action, full=True):
        """ Returns a list already formated to be displayed by
        ``self.main_screen`` as the help for a command.
        """

        action_help = self.get_action_help(action)
        text = [('action-name', action), ' - ',
                '%s' % action_help['short_doc']]

        if full:
            params = _('Parameters')  # noqa f821
            required = _('Required')  # noqa f821

            text += [action_help['doc'], '\n']

            if action_help['parameters']:
                text.append('%s: ' % params)

                for i, param in enumerate(action_help['parameters']):
                    name = param['name']
                    lineinit = ' ' * len(params + ': ') if i != 0 else ''
                    text += [lineinit, ('param', name), ' ']
                    if param['required']:
                        text += [('required', required), '\n']
        text.append('\n')
        return text

    def get_help_screen(self, command_name=None):

        text = [('action-name', 'help'), ' - ']

        text.append(_('Displays this help text.'))  # noqa f821

        params = _('Parameters')  # noqa f821
        text += ['\n%s: ' % params, ('param', 'command-name\n\n')]
        text += [('action-name', 'quit'), ' - ']

        text.append(_('Finishes the program'))  # noqa f821
        text.append('\n')

        ordered_actions = sorted(self.actions.keys())

        for action in ordered_actions:
            text += self.get_action_help_screen(action, full=False)

        return text

    def _get_welcome_text(self):
        # Translators: Do not translate what is inside {}
        return _('Welcome to {toxicbuild}')  # noqa f821

    def _get_help_text(self):
        # Translators: Do not translate what is inside {}
        return _('Type {h} for help and {q} for quit')  # noqa f821

    def _format_help_text(self, text):
        # all this mess to put colors on h and q... pfff
        msg0, msg1 = text.split('{h}')[0], text.split('{h}')[1].split('{q}')[0]
        msg2 = text.split('{h}')[1].split('{q}')[1]

        textstyle = [msg0, ('action-name', 'h'), msg1, ('action-name', 'q'),
                     msg2]
        return textstyle

    def _get_column_sizes(self, output):
        """ Returns a list with the max sizes of for the columns
        of the output. ``output`` is a tuple of tuples, like this:
        `(('a_thing' 'other_thing'), ('value', 'valueb')...)` and returns
        `[7, 11]`.

        """

        sizes = [len(max(c, key=lambda x: len(str(x)))) for c in zip(*output)]
        return sizes

    def _format_result(self, result):
        return str(result)

    def _format_help(self, result):
        return result

    def _format_row(self, sizes, row):
        formated_row = []
        for i, item in enumerate(row):
            item = str(item)
            size = sizes[i]
            itemstr = item + ' ' * (size - len(item)) + ' ' * 4
            formated_row.append(itemstr)

        return ''.join(formated_row)

    def _format_output_columns(self, output):
        sizes = self._get_column_sizes(output)
        formated_output = []
        for row in output:
            row = self._format_row(sizes, row)
            formated_output.append(row)

            formated_output.append('\n')

        return ''.join(formated_output)

    def _format_repo_list(self, repos):
        output = [(_('name'), _('vcs'))]  # noqa f821
        output += [(r['name'], r['vcs_type']) for r in repos]
        return self._format_output_columns(output)

    def _format_slave_list(self, slaves):
        output = [(_('name'), _('host'), _('port'))]  # noqa f821
        output += [(s['name'], s['host'], s['port']) for s in slaves]
        return self._format_output_columns(output)

    def _format_builder_list(self, builders):
        output = [(_('name'), _('status'))]  # noqa f821
        output += [(b['name'], b['status'])
                   for b in builders]
        return self._format_output_columns(output)

    def _format_peek(self, response):
        response = response['body']

        if response['event_type'] in ['build_started', 'build_finished']:
            msg = self._format_peek_build(response)

        elif response['event_type'] in ['step_started', 'step_finished']:
            msg = self._format_peek_step(response)
        else:  # pragma no cover
            msg = ''

        self._peek_output.append(msg)
        return '\n'.join(self._peek_output)

    def _format_peek_build(self, response):
        if response.get('finished'):
            msg = 'build finished with status {}'.format(
                response['status'])

            for step in response['steps']:
                msg += '\n step {} - {}'.format(step['command'],
                                                step['status'])
        else:
            msg = 'build started'

        return msg

    def _format_peek_step(self, response):
        if response.get('finished'):
            # Translators: Do not translate what is inside {}
            msg = _('step {step} finished with status {status}')  # noqa f821
            msg = msg.format(step=response['command'],
                             status=response['status'])
        else:
            # Translators: Do not translate what is inside {}
            msg = _('step {step} is running')  # noqa f821
            msg = msg.format(step=response['command'])

        return msg



# -*- coding: utf-8 -*-

import logging
import os
import pkg_resources
import shutil
import sys
from uuid import uuid4
from toxicbuild.core.conf import Settings
from toxicbuild.core.cmd import command, main
from toxicbuild.core.utils import (daemonize as daemon, bcrypt_string, bcrypt,
                                   changedir)


ENVVAR = 'TOXICSLAVE_SETTINGS'
DEFAULT_SETTINGS = 'toxicslave.conf'

settings = None

PIDFILE = 'toxicslave.pid'
LOGFILE = 'toxicslave.log'


def create_settings():
    global settings

    settings = Settings(ENVVAR, DEFAULT_SETTINGS)


@command
def start(workdir, daemonize=False, stdout=LOGFILE,
          stderr=LOGFILE, conffile=None, loglevel='info',
          pidfile=PIDFILE):
    """ Starts toxicslave.

    Starts the build server to listen on the specified port for
    requests from addr (0.0.0.0 means everyone). Addr and port params
    came from the config file

    :param workdir: Work directory for server.
    :param --daemonize: Run as daemon. Defaults to False
    :param --stdout: stdout path. Defaults to /dev/null
    :param --stderr: stderr path. Defaults to /dev/null
    :param -c, --conffile: path to config file. Defaults to None.
      If not conffile, will look for a file called ``toxicslave.conf``
      inside ``workdir``
    :param --loglevel: Level for logging messages. Defaults to `info`.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
      ``toxicslave.pid``
    """

    print('Starting toxicslave')
    if not os.path.exists(workdir):
        print('Workdir `{}` does not exist'.format(workdir))
        sys.exit(1)

    if conffile:
        os.environ['TOXICSLAVE_SETTINGS'] = conffile
    else:
        os.environ['TOXICSLAVE_SETTINGS'] = os.path.join(workdir,
                                                         'toxicslave.conf')

    create_settings()
    # These toxicbuild.slave imports must be here so I can
    # change the settings file before settings are instanciated.
    from toxicbuild.slave import server, settings

    addr = settings.ADDR
    port = settings.PORT

    if daemonize:
        daemon(call=server.run_server, cargs=(addr, port), ckwargs={},
               stdout=stdout, stderr=stderr, workdir=workdir, pidfile=pidfile)
    else:
        loglevel = getattr(logging, loglevel.upper())
        logging.basicConfig(level=loglevel)

        with changedir(workdir):
            server.run_server(addr, port)


@command
def stop(workdir, pidfile=PIDFILE):
    """ Stops toxicslave.

    The instance of toxicslave in ``workdir`` will be stopped.

    :param workdir: Workdir for master to be killed.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
      ``toxicslave.pid``
    """

    print('Stopping toxicslave')
    with changedir(workdir):
        with open(pidfile) as fd:
            pid = int(fd.read())

        os.kill(pid, 9)
        os.remove(pidfile)


@command
def restart(workdir, pidfile=PIDFILE):
    """Restarts toxicslave

    The instance of toxicslave in ``workdir`` will be restarted.

    :param workdir: Workdir for master to be killed.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
        ``toxicslave.pid``
    """

    stop(workdir, pidfile=pidfile)
    start(workdir, pidfile=pidfile, daemonize=True)


@command
def create(root_dir):
    """ Create a new toxicslave environment.

    :param --root_dir: Root directory for toxicslave.
    """
    print('Creating root_dir `{}` for toxicslave'.format(root_dir))

    # First we create the directory
    os.makedirs(root_dir)

    # after that we copy the config file to the root dir
    template_fname = 'toxicslave.conf.tmpl'
    template_dir = pkg_resources.resource_filename('toxicbuild.slave',
                                                   'templates')
    template_file = os.path.join(template_dir, template_fname)
    dest_file = os.path.join(root_dir, 'toxicslave.conf')
    shutil.copyfile(template_file, dest_file)

    # here we create a bcrypt salt and a access token for authentication.
    salt = bcrypt.gensalt(8)
    access_token = str(uuid4())
    encrypted_token = bcrypt_string(access_token, salt)

    # and finally update the config file content with the new generated
    # salt and access token
    with open(dest_file, 'r+') as fd:
        content = fd.read()
        content = content.replace('{{BCRYPT_SALT}}', salt.decode())
        content = content.replace('{{ACCESS_TOKEN}}', encrypted_token)
        fd.seek(0)
        fd.write(content)

    print('Toxicslave environment created with access token: {}'.format(
        access_token))
    return access_token


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright 2015-2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
from collections import defaultdict
import os
from toxicbuild.core import get_vcs
from toxicbuild.core.utils import (get_toxicbuildconf, LoggerMixin,
                                   ExecCmdError, match_string)
from toxicbuild.slave import settings
from toxicbuild.slave.build import Builder, BuildStep
from toxicbuild.slave.docker import DockerContainerBuilder
from toxicbuild.slave.exceptions import (BuilderNotFound, BadBuilderConfig,
                                         BusyRepository)
from toxicbuild.slave.plugins import SlavePlugin


class BuildManager(LoggerMixin):

    """ A manager for remote build requests
    """

    # repositories that are being cloned
    cloning_repos = set()
    # repositories that are being updated
    updating_repos = set()
    # repositories that are building something.
    # key is repo_url and value is named_tree
    building_repos = defaultdict(lambda: None)  # pragma no branch WTF??

    def __init__(self, protocol, repo_url, vcs_type, branch, named_tree):
        self.protocol = protocol
        self.repo_url = repo_url
        self.vcs_type = vcs_type
        self.vcs = get_vcs(vcs_type)(self.workdir)
        self.branch = branch
        self.named_tree = named_tree
        self._configmodule = None

    def __enter__(self):
        if self.current_build and self.current_build != self.named_tree:
            msg = '{} is busy at {}. Can\'t work at {}'.format(
                self.repo_url, self.current_build, self.named_tree)
            raise BusyRepository(msg)

        self.current_build = self.named_tree
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.current_build = None

    @property
    def configmodule(self):
        if not self._configmodule:  # pragma: no branch
            self._configmodule = get_toxicbuildconf(self.workdir)
        return self._configmodule

    @property
    def workdir(self):
        """ The directory where the source code of this repository is
        cloned into
        """
        workdir = self.repo_url.replace('/', '-').replace('@', '').replace(
            ':', '')
        return os.path.join('src', workdir)

    @property
    def current_build(self):
        return type(self).building_repos.get(self.repo_url)

    @current_build.setter
    def current_build(self, value):
        type(self).building_repos[self.repo_url] = value

    @property
    def is_cloning(self):
        """Informs if this repository is being cloned."""

        return self.repo_url in type(self).cloning_repos

    @is_cloning.setter
    def is_cloning(self, value):
        if value is True:
            type(self).cloning_repos.add(self.repo_url)
        else:
            type(self).cloning_repos.discard(self.repo_url)

    @property
    def is_updating(self):
        """Informs it this repository is fetching changes"""
        return self.repo_url in type(self).updating_repos

    @is_updating.setter
    def is_updating(self, value):
        if value is True:
            type(self).updating_repos.add(self.repo_url)
        else:
            type(self).updating_repos.discard(self.repo_url)

    @property
    def is_working(self):
        """Informs if this repository is cloning or updating"""
        return self.is_cloning or self.is_updating

    @asyncio.coroutine
    def wait_clone(self):
        """Wait until the repository clone is complete."""

        while self.is_cloning:
            yield from asyncio.sleep(1)

    @asyncio.coroutine
    def wait_update(self):
        """Wait until the repository update is complete."""

        while self.is_updating:
            yield from asyncio.sleep(1)

    @asyncio.coroutine
    def wait_all(self):
        """Wait until clone and update are done."""

        while self.is_working:
            yield from asyncio.sleep(1)

    @asyncio.coroutine
    def update_and_checkout(self, work_after_wait=True):
        """ Updates ``self.branch`` and checkout to ``self.named_tree``.
        :param work_after_wait: Indicates if we should update and checkout
          after waiting for other instance finishes its job.
        """

        if self.is_working:
            yield from self.wait_all()
            if not work_after_wait:
                return

        try:
            self.is_updating = True
            if not self.vcs.workdir_exists():
                self.log('cloning {}'.format(self.repo_url))
                yield from self.vcs.clone(self.repo_url)

            # first we try to checkout to the named_tree because if if
            # already exists here we don't need to update the code.
            try:
                self.log('checking out to named_tree {}'.format(
                    self.named_tree), level='debug')
                yield from self.vcs.checkout(self.named_tree)
            except ExecCmdError:
                # this is executed when the named_tree does not  exist
                # so we upate the code and then checkout again.
                self.log('named_tree does not exist. updating...')
                self.log('checking out to branch {}'.format(self.branch),
                         level='debug')
                yield from self.vcs.checkout(self.branch)
                yield from self.vcs.pull(self.branch)
                self.log('checking out to named_tree {}'.format(
                    self.named_tree), level='debug')
                yield from self.vcs.checkout(self.named_tree)

        finally:
            self.is_updating = False

    def _branch_match(self, builder):
        return builder.get('branch') is None or match_string(
            self.branch, [builder.get('branch', '')])

        # the whole purpose of toxicbuild is this!
        # see the git history and look for the first versions.
        # First thing I changed on buildbot was to add the possibility
        # to load builers from a config file.
    def list_builders(self):
        """ Returns a list with all builders names for this branch
        based on toxicbuild.conf file
        """

        try:
            builders = [b['name'] for b in self.configmodule.BUILDERS
                        if (b.get('branch') == self.branch or b.get(
                            'branch')is None)]
        except KeyError as e:
            key = str(e)
            msg = 'Your builder config does not have a required key'.format(
                key)
            raise BadBuilderConfig(msg)

        return builders

    async def load_builder(self, name):
        """ Load a builder from toxicbuild.conf. If a container
        is to be used in for the build, returns a container builder
        instance. Otherwise, return a Builder instance.

        :param name: builder name
        """
        try:
            bdict = [b for b in self.configmodule.BUILDERS if (b.get(
                'branch') is None and b['name'] == name) or (b.get(
                    'branch') == self.branch and b['name'] == name)][0]
        except IndexError:
            msg = 'builder {} does not exist for {} branch {}'.format(
                name, self.repo_url, self.branch)
            raise BuilderNotFound(msg)

        platform = bdict.get('platform', 'linux-generic')
        remove_env = bdict.get('remove_env', True)
        # now we have all we need to instanciate the container builder if
        # needed.
        if settings.USE_DOCKER:
            builder = DockerContainerBuilder(self, platform, self.repo_url,
                                             self.vcs_type, self.branch,
                                             self.named_tree,
                                             name, self.workdir,
                                             remove_env=remove_env)
        else:
            # this envvars are used in all steps in this builder
            builder_envvars = bdict.get('envvars', {})
            builder = Builder(self, bdict['name'], self.workdir, platform,
                              remove_env=remove_env, **builder_envvars)
            builder.steps = await self._get_builder_steps(builder, bdict)

        return builder

    async def _get_builder_steps(self, builder, bdict):
        plugins_conf = bdict.get('plugins')
        steps = []

        if plugins_conf:
            builder.plugins = await self._load_plugins(plugins_conf)

        for plugin in builder.plugins:
            steps += plugin.get_steps_before()

        for sdict in bdict['steps']:
            step = BuildStep(**sdict)
            steps.append(step)

        for plugin in builder.plugins:
            steps += plugin.get_steps_after()

        return steps

    # kind of wierd place for this thing
    @asyncio.coroutine
    def send_info(self, info):
        yield from self.protocol.send_response(code=0, body=info)

    def log(self, msg, level='info'):
        msg = '[{}]{}'.format(self.repo_url, msg)
        super().log(msg, level=level)

    async def _load_plugins(self, plugins_config):
        """ Returns a list of :class:`toxicbuild.slave.plugins.Plugin`
        subclasses based on the plugins listed on the config for a builder.
        """

        plist = []
        for pdict in plugins_config:
            plugin_class = SlavePlugin.get_plugin(pdict['name'])
            del pdict['name']
            plugin = plugin_class(**pdict)
            if getattr(plugin, 'uses_data_dir', False):
                await plugin.create_data_dir()
            plist.append(plugin)
        return plist



# -*- coding: utf-8 -*-

import asyncio
try:
    from asyncio import ensure_future
except ImportError:
    from asyncio import async as ensure_future

import logging
import os
import pkg_resources
import shutil
import sys
from uuid import uuid4
from mongomotor import connect
from toxicbuild.core.cmd import command, main
from toxicbuild.core.conf import Settings
from toxicbuild.core.utils import (log, daemonize as daemon, bcrypt,
                                   bcrypt_string, changedir)
from toxicbuild.master.scheduler import TaskScheduler

settings = None
dbconn = None
scheduler = None


PIDFILE = 'toxicmaster.pid'
LOGFILE = 'toxicmaster.log'

ENVVAR = 'TOXICMASTER_SETTINGS'
DEFAULT_SETTINGS = 'toxicmaster.conf'


def create_settings_and_connect():
    global settings, dbconn

    settings = Settings(ENVVAR, DEFAULT_SETTINGS)
    # here the database connection
    dbsettings = settings.DATABASE
    dbconn = connect(**dbsettings)


def ensure_indexes(*classes):
    for cls in classes:
        cls.ensure_indexes()


def create_scheduler():
    global scheduler

    scheduler = TaskScheduler()
    ensure_future(scheduler.start())


@asyncio.coroutine
def toxicinit():
    """ Initialize services. """

    create_settings_and_connect()

    # importing here to avoid circular imports
    from toxicbuild.master.build import BuildSet
    from toxicbuild.master.hole import HoleServer
    from toxicbuild.master.repository import Repository
    from toxicbuild.master.slave import Slave
    from toxicbuild.master.users import User, Organization

    ensure_indexes(BuildSet, Repository, Slave, User, Organization)
    create_scheduler()

    log('[init] Scheduling all')
    yield from Repository.schedule_all()
    if settings.ENABLE_HOLE:
        hole_host = settings.HOLE_ADDR
        hole_port = settings.HOLE_PORT
        server = HoleServer(hole_host, hole_port)
        log('[init] Serving UIHole at {}'.format(settings.HOLE_PORT))
        server.serve()

    log('[init] Toxicbuild is running!')


def run(loglevel):
    """ Runs Toxicbuild master """

    loglevel = getattr(logging, loglevel.upper())
    logging.basicConfig(level=loglevel)

    loop = asyncio.get_event_loop()
    loop.run_until_complete(toxicinit())
    loop.run_forever()


# console commands


@command
def start(workdir, daemonize=False, stdout=LOGFILE, stderr=LOGFILE,
          conffile=None, loglevel='info', pidfile=PIDFILE):
    """ Starts toxicmaster.

    :param workdir: Work directory for server.
    :param --daemonize: Run as daemon. Defaults to False
    :param --stdout: stdout path. Defaults to /dev/null
    :param --stderr: stderr path. Defaults to /dev/null
    :param -c, --conffile: path to config file. Defaults to None.
      If not conffile, will look for a file called ``toxicmaster.conf``
      inside ``workdir``
    :param --loglevel: Level for logging messages. Defaults to `info`.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
      ``toxicslave.pid``
    """

    print('Starting toxicmaster')
    if not os.path.exists(workdir):
        print('Workdir `{}` does not exist'.format(workdir))
        sys.exit(1)

    if conffile:
        os.environ['TOXICMASTER_SETTINGS'] = conffile
    else:
        os.environ['TOXICMASTER_SETTINGS'] = DEFAULT_SETTINGS

    if daemonize:
        daemon(call=run, cargs=(loglevel,), ckwargs={}, stdout=stdout,
               stderr=stderr, workdir=workdir, pidfile=pidfile)
    else:
        with changedir(workdir):
            run(loglevel)


@command
def stop(workdir, pidfile=PIDFILE):
    """ Kills toxicmaster.

    :param --workdir: Workdir for master to be killed. Looks for a file
      ``toxicmaster.pid`` inside ``workdir``.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
      ``toxicslave.pid``
    """

    print('Stopping toxicmaster')
    with changedir(workdir):
        with open(pidfile) as fd:
            pid = int(fd.read())

        os.kill(pid, 9)
        os.remove(pidfile)


@command
def restart(workdir, pidfile=PIDFILE):
    """Restarts toxicmaster

    The instance of toxicmaster in ``workdir`` will be restarted.
    :param workdir: Workdir for master to be killed.
    :param --pidfile: Name of the file to use as pidfile.  Defaults to
    ``toxicmaster.pid``
    """

    stop(workdir, pidfile=pidfile)
    start(workdir, pidfile=pidfile, daemonize=True)


@command
def create(root_dir):
    """ Creates a new toxicmaster environment.

    :param --root_dir: Root directory for toxicmaster.
    """
    print('Creating root_dir `{}` for toxicmaster'.format(root_dir))

    os.makedirs(root_dir)

    template_fname = 'toxicmaster.conf.tmpl'
    template_dir = pkg_resources.resource_filename('toxicbuild.master',
                                                   'templates')
    template_file = os.path.join(template_dir, template_fname)
    dest_file = os.path.join(root_dir, 'toxicmaster.conf')
    shutil.copyfile(template_file, dest_file)

    # here we create a bcrypt salt and a access token for authentication.
    salt = bcrypt.gensalt(8)
    access_token = str(uuid4())
    encrypted_token = bcrypt_string(access_token, salt)

    # and finally update the config file content with the new generated
    # salt and access token
    with open(dest_file, 'r+') as fd:
        content = fd.read()
        content = content.replace('{{BCRYPT_SALT}}', salt.decode())
        content = content.replace('{{ACCESS_TOKEN}}', encrypted_token)
        fd.seek(0)
        fd.write(content)

    print('Toxicmaster environment created with access token: {}'.format(
        access_token))

    return access_token


@command
def create_user(configfile, email=None, password=None, superuser=False):
    """Creates a superuser in the master.

    :param --email: User's email.
    :param --password: Password for authentication.
    :param --superuser: Indicates if the user is a super user.
      Defaults to False"""

    print('Creating user for authenticated access')

    os.environ[ENVVAR] = configfile
    create_settings_and_connect()

    from toxicbuild.master.users import User  # noqa f401

    if not email:
        email = _ask_thing('email: ')

    if not password:
        password = _ask_thing('password: ')

    user = User(email=email, is_superuser=superuser,
                allowed_actions=['add_repo', 'add_slave', 'remove_user'])
    user.set_password(password)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(user.save())
    print('User {} created successfully'.format(user.username))


def _ask_thing(thing, opts=[]):
    if opts:
        thing += '[' + '/'.join(opts) + ']'

    response = input(thing)
    while not response:
        response = input(thing)
        if opts:
            if response.lower() not in opts:
                response = ''

    return response


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

# This is a very sketchy thing. We use a set of pre built images (they are
# mapped in the settings file) and each time we run a build we run a
# container based in one of these images.

# The oddest part is that when the container is done we `docker cp` - this is
# why the silly CONTAINER_SLAVE_WORKDIR settings variable exists - the
# source code to the container instead of creating an image with the source
# code.

# This is odd, but there is a reason for that. The files are
# copied to the container instead of create a container with the files
# because I want to have the option to preserve the environment for future
# builds.

import asyncio
from toxicbuild.core.exceptions import ConfigError
from toxicbuild.core.utils import exec_cmd, LoggerMixin
from toxicbuild.slave import settings
from toxicbuild.slave.client import ContainerBuildClient


class DockerContainerBuilder(LoggerMixin):
    """Class to handle docker containers used to run builds inside it.
    """

    # really ugly signature, right?
    def __init__(self, manager, platform, repo_url, vcs_type, branch,
                 named_tree, builder_name, source_dir, remove_env=True):
        """Constructor for DockerContainer.

        :param manager: instance of :class:`toxicbuild.slave.BuildManager.`
        :param platform: Platform where the build will be executed
        :param repo_url: The repository URL
        :param vcs_type: Type of vcs used in the repository.
        :param branch: Which branch to use in the build.
        :param named_tree: A tag, commit, branch name...
        :param builder_name: The builder that will execute the build
        :param source_dir: Directory with the source code in the host
        :param remove_env: Should the container be removed when the build is
          done?"""

        self.build_data_body = {'repo_url': repo_url,
                                'vcs_type': vcs_type,
                                'branch': branch,
                                'named_tree': named_tree,
                                'builder_name': builder_name}
        self.build_data = {'action': 'build',
                           'body': self.build_data_body}

        self.manager = manager
        self.source_dir = source_dir
        self.builder_name = builder_name
        self.container_slave_workdir = settings.CONTAINER_SLAVE_WORKDIR
        self.platform = platform
        self.image_name = settings.DOCKER_IMAGES[self.platform]
        self.remove_env = remove_env
        self.docker_cmd = 'docker'
        self.client = ContainerBuildClient(self.build_data, self)

    @property
    def name(self):
        name = '{}-{}-{}'.format(
            self.source_dir.replace('/', '-').replace('\\', '-'),
            self.platform, self.builder_name)
        return name

    async def __aenter__(self):
        await self.start_container()
        await self.copy2container()
        await self.client.__aenter__()
        return self

    async def __aexit__(self, ext_typ, exc_val, exc_tb):
        await self.client.__aexit__(ext_typ, exc_val, exc_tb)
        if not self.remove_env:
            # removes the source code from the container
            await self.rm_from_container()

        await self.kill_container()

        if self.remove_env:
            await self.rm_container()

    async def container_exists(self):
        """Checks if a container named as its ``self.name``
        already exists"""

        msg = 'Checking if container exists'
        self.log(msg, level='debug')
        cmd = '{} container ps -a --filter name={}'.format(self.docker_cmd,
                                                           self.name)
        cmd += '| wc -l'
        msg = 'Executing {}'.format(cmd)
        self.log(msg, level='debug')
        ret = await exec_cmd(cmd, cwd='.')
        return int(ret) > 1

    async def is_running(self):
        try:
            async with self.client:
                is_running = await self.client.healthcheck()
        except Exception:
            is_running = False
        return is_running

    async def wait_start(self):
        msg = 'Waiting slave to start'
        self.log(msg, level='debug')
        try:
            timeout = settings.DOCKER_CONTAINER_TIMEOUT
        except ConfigError:
            timeout = 30
        total = 0.0
        step = 0.1
        is_running = await self.is_running()
        while not is_running and total < timeout:
            await asyncio.sleep(step)
            total += step
            is_running = await self.is_running()

    async def start_container(self):
        exists = await self.container_exists()
        self.log('Starting container {}'.format(self.name),
                 level='debug')
        if not exists:
            cmd = '{} run -t -d --name {} {}'.format(self.docker_cmd,
                                                     self.name,
                                                     self.image_name)

        else:
            cmd = '{} start {}'.format(self.docker_cmd, self.name)

        self.log(cmd, level='debug')
        await exec_cmd(cmd, cwd='.')
        await self.wait_start()

    async def kill_container(self):
        msg = 'Killing container {}'.format(self.name)
        self.log(msg, level='debug')
        cmd = '{} kill {}'.format(self.docker_cmd, self.name)
        self.log(cmd, level='debug')
        await exec_cmd(cmd, cwd='.')

    async def rm_container(self):
        msg = 'Removing container {}'.format(self.name)
        self.log(msg, level='debug')
        cmd = '{} rm {}'.format(self.docker_cmd, self.name)
        self.log(cmd, level='debug')
        await exec_cmd(cmd, cwd='.')

    async def copy2container(self):
        """Recursive copy a directory to the container's src dir."""

        msg = 'Copying files to container {}'.format(self.name)
        self.log(msg, level='debug')
        cmd = 'cd {} && {} cp . {}:{}/{}/'.format(
            self.source_dir, self.docker_cmd, self.name,
            self.container_slave_workdir, self.source_dir)
        self.log(cmd, level='debug')
        await exec_cmd(cmd, cwd='.')

    async def rm_from_container(self):
        """Removes the source code of a container that will not be removed.
        """

        msg = 'Removing files from container {}'.format(self.name)
        self.log(msg, level='debug')
        # removing source dir
        cmd = '{} exec {} rm -rf {}/{}'.format(self.docker_cmd, self.name,
                                               self.container_slave_workdir,
                                               self.source_dir)
        msg = 'Executing {}'.format(cmd)
        self.log(msg, level='debug')
        await exec_cmd(cmd, cwd='.')

        # removing build dir
        cmd = '{} exec {} rm -rf {}/{}-{}'.format(self.docker_cmd, self.name,
                                                  self.container_slave_workdir,
                                                  self.source_dir,
                                                  self.builder_name)
        msg = 'Executing {}'.format(cmd)
        self.log(msg, level='debug')
        await exec_cmd(cmd, cwd='.')

    async def get_container_ip(self):
        # look what a beautiful command!
        msg = 'Getting container ip'
        self.log(msg, level='debug')
        ip_regex = '((1?[0-9][0-9]?|2[0-4][0-9]|25[0-5])\.){3}'
        ip_regex += '(1?[0-9][0-9]?|2[0-4][0-9]|25[0-5])'
        cmd = [self.docker_cmd, 'inspect', self.name, '|', 'grep', '-i',
               'IPAddress', '|', 'grep', '-oE', '-m', '1',
               "'{}'".format(ip_regex)]
        cmd = ' '.join(cmd)
        msg = 'Executing {}'.format(cmd)
        self.log(msg, level='debug')
        r = await exec_cmd(cmd, cwd='.')
        return r

    async def build(self):
        async with self:
            await self.client.build(self.manager.protocol.send_response)



# -*- coding: utf-8 -*-

# Copyright 2015 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from abc import ABCMeta, abstractmethod
import asyncio
import os
from toxicbuild.core.exceptions import VCSError
from toxicbuild.core.utils import (exec_cmd, inherit_docs, string2datetime,
                                   datetime2string, utc2localtime,
                                   localtime2utc, LoggerMixin, match_string)


class VCS(LoggerMixin, metaclass=ABCMeta):

    """ Generic inteface to a vcs (clone, fetch, get revisions etc...).
    """
    vcsbin = None

    def __init__(self, workdir):
        """:param workdir: Directory where repository will be cloned and
        all action will happen.
        """
        self.workdir = workdir

    @asyncio.coroutine
    def exec_cmd(self, cmd, cwd=None):
        """ Executes a shell command. If ``cwd`` is None ``self.workdir``
        will be used.

        :param cwd: Directory where the command will be executed.
        """
        if cwd is None:
            cwd = self.workdir

        ret = yield from exec_cmd(cmd, cwd)
        return ret

    def workdir_exists(self):
        """ Informs if the workdir for this vcs exists
        """
        return os.path.exists(self.workdir)

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def clone(self, url):
        """ Clones a repository into ``self.workdir``
        :param url: repository url
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def fetch(self):
        """ Fetch changes from remote repository
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def checkout(self, named_tree):
        """ Checkout to ``named_tree``
        :param named_tree: A commit, branch, tag...
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def pull(self, branch_name):
        """ Pull changes from ``branch_name`` on remote repo.

        :param branch_name: A branch name, like 'master'.
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def has_changes(self):
        """ Informs if there are new revisions in the repository
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def get_revisions(self, since={}, branches=None):
        """ Returns the revisions for ``branches`` since ``since``.

        :param since: dictionary in the format: {branch_name: since_date}.
          ``since`` is a datetime object.
        :param branches: A list of branches to look for new revisions. If
          ``branches`` is None all remote branches will be used. You can use
          wildcards in branches to filter the remote branches.
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def get_revisions_for_branch(self, branch, since=None):
        """ Returns the revisions for ``branch`` since ``since``.
        If ``since`` is None, all revisions will be returned.

        :param branch: branch name
        :param since: datetime
        """

    @abstractmethod  # pragma no branch
    @asyncio.coroutine
    def get_remote_branches(self):
        """ Returns a list of the remote branches available.
        """

    def _filter_remote_branches(self, remote_branches, branch_filters):
        """Filters the remote branches based in filters for the branches'
        names."""

        return [b for b in remote_branches if match_string(b, branch_filters)]


@inherit_docs
class Git(VCS):

    """ An interface to git version control system
    """

    vcsbin = 'git'
    # this date_format is used to ask git about revisions since
    # some date
    date_format = '%a %b %d %H:%M:%S %Y'

    @asyncio.coroutine
    def clone(self, url):

        cmd = '%s clone %s %s --recursive' % (self.vcsbin, url, self.workdir)
        # we can't go to self.workdir while we do not clone the repo
        yield from self.exec_cmd(cmd, cwd='.')

    @asyncio.coroutine
    def fetch(self):
        cmd = '%s %s' % (self.vcsbin, 'fetch')

        fetched = yield from self.exec_cmd(cmd)
        return fetched

    @asyncio.coroutine
    def checkout(self, named_tree):

        cmd = '{} checkout {}'.format(self.vcsbin, named_tree)
        yield from self.exec_cmd(cmd)

    @asyncio.coroutine
    def pull(self, branch_name):

        cmd = '{} pull --no-edit origin {}'.format(self.vcsbin, branch_name)

        ret = yield from self.exec_cmd(cmd)
        return ret

    @asyncio.coroutine
    def has_changes(self):
        ret = yield from self.fetch()
        return bool(ret)

    @asyncio.coroutine
    def update_submodule(self):
        cmd = '{} submodule init'.format(self.vcsbin)
        yield from self.exec_cmd(cmd)
        cmd = '{} submodule update'.format(self.vcsbin)
        ret = yield from self.exec_cmd(cmd)
        return ret

    @asyncio.coroutine
    def get_revisions(self, since={}, branches=None):

        # this must be called everytime so we sync our repo
        # with the remote repo and then we can see new branches
        yield from self.fetch()
        remote_branches = yield from self.get_remote_branches()
        if branches:
            remote_branches = self._filter_remote_branches(
                remote_branches, branches)

        revisions = {}
        for branch in remote_branches:
            try:
                yield from self.checkout(branch)
                yield from self.pull(branch)
                since_date = since.get(branch)

                revs = yield from self.get_revisions_for_branch(branch,
                                                                since_date)
                revisions[branch] = revs
            except Exception as e:
                msg = 'Error fetching changes. {}'.format(str(e))
                self.log(msg)

        return revisions

    @asyncio.coroutine
    def get_revisions_for_branch(self, branch, since=None):

        # hash | commit date | author | title
        cmd = '{} log --pretty=format:"%H | %ad | %an | %s" '.format(
            self.vcsbin)
        if since:
            # Here we change the time to localtime since we can't get
            # utc time in git commits unless we are using git 2.7+
            localtime = utc2localtime(since)
            date = datetime2string(localtime, self.date_format)
            self.log('get revisions for branch {} since {}'.format(branch,
                                                                   date),
                     level='debug')
            cmd += '--since="%s" ' % date

        cmd += '--date=local'
        last_revs = [r for r in (yield from self.exec_cmd(cmd)).split('\n')
                     if r]

        last_revs.reverse()
        revisions = []

        for rev in last_revs:
            rev_uuid, date, author, title = rev.split(' | ')
            date = string2datetime(date.strip(), dtformat=self.date_format)
            # Here we change the date from git, that is in localtime to
            # utc before saving to database.
            date = localtime2utc(date)
            revisions.append({'commit': rev_uuid.strip(), 'commit_date': date,
                              'author': author, 'title': title})

        # The thing here is that the first revision in the list
        # is the last one consumed on last time
        return revisions[1:]

    @asyncio.coroutine
    def get_remote_branches(self):
        yield from self.fetch()
        yield from self._update_remote_prune()
        cmd = '%s branch -r' % self.vcsbin

        out = yield from self.exec_cmd(cmd)
        msg = 'Remote branches: {}'.format(out)
        self.log(msg, level='debug')
        remote_branches = out.split('\n')
        # master, with some shitty arrow...
        remote_branches[0] = remote_branches[0].split('->')[1].strip()
        return set([b.strip().split('/')[1] for b in remote_branches])

    @asyncio.coroutine
    def _update_remote_prune(self):
        """Updates remote branches list, prunning deleted branches."""

        cmd = '{} remote update --prune'.format(self.vcsbin)
        msg = 'Updating --prune remote'
        self.log(msg, level='debug')
        yield from self.exec_cmd(cmd)


VCS_TYPES = {'git': Git}


def get_vcs(vcs_type):
    """ Retuns a subclass of :class:`toxicbuild.core.vcs.VCS` for ``vcs_type``
    """
    vcs = VCS_TYPES.get(vcs_type)
    if not vcs:
        raise VCSError('VCS not found for {}'.format(vcs_type))
    return vcs



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

# !!!!!!
# after user stuff was implemented, this whole thing is useless now.
# It must be removed.
# !!!!!!

import asyncio
try:
    from asyncio import async as ensure_future
except ImportError:  # pragma no cover
    from asyncio import ensure_future

from asyncblink import signal
from toxicbuild.core.utils import LoggerMixin
from toxicbuild.ui import settings
from toxicbuild.ui.client import get_hole_client


message_arrived = signal('message-arrived')


class StreamConnector(LoggerMixin):
    """Class responsible for connecting the ui with the master's stream.
    It is done so we don't have one connection open for each client.

    We have a client for each repository and the clients register callables
    to receive the messages sent throught the stream. To use it, simply
    plug a callback to the connector:

    .. code-block:: python

       from toxicbuild.ui.models import Repository
       repo = yield from Repository.get(id='some-id')

       def callback(sender, **message):
           print(message)

       StreamConnector.plug(repo, callback)

    When you are done, unplug it.

    .. code-block:: python

       StreamConnector.unplug(repo, callback)

    """

    NONE_REPO_ID = 'NONE-REPO-ID'
    # limit is not implemented
    _clients_limit = 10
    _instances = {}

    def __init__(self, user, repo_id):
        self.clients_connected = 0
        self.client = None
        self.repo_id = repo_id
        self.user = user
        self._connected = False

    @asyncio.coroutine
    def _connect(self):
        if self._connected:
            self.log('Client already connected', level='warning')
            return

        host = settings.HOLE_HOST
        port = settings.HOLE_PORT
        client = yield from get_hole_client(self.user, host, port)
        yield from client.connect2stream()
        self.client = client
        self._connected = True

    def _disconnect(self):
        self.client.disconnect()
        self._connected = False

    def _get_repo_id(self, body):
        if 'build' in body.keys():
            return body.get('build').get('repository').get('id')
        return body.get('repository', {}).get('id')

    @asyncio.coroutine
    def _listen(self):
        yield from self._connect()
        while self._connected:
            response = yield from self.client.get_response()
            body = response.get('body')
            if body is None:
                self.log('Bad data from stream. Skipping',
                         level='warning')
                break

            repo_id = self._get_repo_id(body)
            event = body.get('event_type')
            if repo_id == self.repo_id or event == 'repo_status_changed':
                message_arrived.send(repo_id, **body)

    @classmethod
    @asyncio.coroutine
    def _prepare_instance(cls, user, repo_id):
        """Returns an instance of
        :class:`toxicbuild.ui.connectors.StreamConnector` that will
        notify about messages from a specific repository.

        :param repo_id: Id of a repository. If repo_id is None, a default
          :attr:`toxicbuild.ui.connectors.StreamConnector` will be used as
          the key."""

        if repo_id is None:
            repo_id = cls.NONE_REPO_ID

        inst = cls._instances.get((user.id, repo_id))
        if not inst:
            inst = cls(user, repo_id)
            cls._instances[(user.id, repo_id)] = inst
            ensure_future(inst._listen())

        inst.clients_connected += 1
        return inst

    @classmethod
    def _release_instance(cls, user, repo_id):

        if repo_id is None:
            repo_id = cls.NONE_REPO_ID

        conn = cls._instances[(user.id, repo_id)]
        conn.clients_connected -= 1

        if conn.clients_connected < 1:
            conn._disconnect()
            cls._instances.pop((user.id, repo_id))

    @classmethod
    @asyncio.coroutine
    def plug(cls, user, repo_id, callback):
        """Connects ``callback`` to events sent by a repository.

        :param user: The requester user.
        :param repo_id: The Id of a repository. Messages sent by this
          repository will trigger ``callback``. If repo_id is None, messages
          from all repositories will be sent to ``callback``.
        :param callback: A callable that will handle messages from
          a repository."""

        yield from cls._prepare_instance(user, repo_id)
        kw = {}
        if repo_id is not None:
            kw = {'sender': repo_id}

        message_arrived.connect(callback, **kw)

    @classmethod
    def unplug(cls, user, repo_id, callback):
        """Disconnects ``callback`` from events sent by a repository.

        :param user: The requester user.
        :param repo_id: The id of a repository.
        :param callback: A callable that will handle messages from
          a repository."""

        message_arrived.disconnect(callback, sender=repo_id)
        cls._release_instance(user, repo_id)



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

# THIS WHOLE MODULE NEEDS TO BE RE-WRITTEN

import asyncio
from asyncio import ensure_future
import base64
import datetime
import json
import traceback

from tornado import gen
from tornado.websocket import WebSocketHandler, WebSocketError

from pyrocumulus.web.applications import (PyroApplication, StaticApplication)
from pyrocumulus.web.handlers import TemplateHandler, PyroRequest
from pyrocumulus.web.urlmappers import URLSpec

from toxicbuild.core.utils import LoggerMixin, string2datetime
from toxicbuild.ui.connectors import StreamConnector
from toxicbuild.ui.models import (Repository, Slave, BuildSet, Builder, Plugin,
                                  User)
from toxicbuild.ui.utils import format_datetime, is_datetime


COOKIE_NAME = 'toxicui'


class LoginHandler(TemplateHandler):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None

    def get(self, action):
        if action == 'logout':
            self.clear_cookie(COOKIE_NAME)
            return self.redirect('/')

        if self.get_secure_cookie(COOKIE_NAME):
            return self.redirect('/')

        error = bool(self.params.get('error'))
        self.render_template('login.html', {'error': error})

    @gen.coroutine
    def post(self, action):
        username_or_email = self.params.get('username_or_email')
        password = self.params.get('password')

        if not (username_or_email and password):
            return self.redirect('/login?error=2')

        try:
            self.user = yield from User.authenticate(username_or_email,
                                                     password)
        except:
            return self.redirect('/login?error=1')

        self._set_cookie_content()
        self.redirect('/')

    def _set_cookie_content(self):
        userjson = json.dumps({'id': self.user.id, 'email': self.user.email,
                               'username': self.user.username})

        content = base64.encodebytes(userjson.encode('utf-8'))
        self.set_secure_cookie(COOKIE_NAME, content)


class LoggedTemplateHandler(TemplateHandler):

    def prepare(self):
        super().prepare()
        user = self._get_user_from_cookie()
        if not user:
            self.redirect('/login')
        self.user = user

    def _get_user_from_cookie(self):
        cookie = self.get_secure_cookie(COOKIE_NAME)
        if not cookie:
            return

        userjson = base64.decodebytes(cookie).decode('utf-8')
        return User(None, json.loads(userjson))


class BaseModelHandler(LoggedTemplateHandler):
    item_template = 'item.html'
    list_template = 'list.html'

    def initialize(self, *args, **kwargs):
        self.model = kwargs['model']
        del kwargs['model']
        super().initialize(*args, **kwargs)

    @asyncio.coroutine
    def get_item(self, **kwargs):
        item = yield from self.model.get(self.user, **kwargs)
        return item

    @asyncio.coroutine
    def add(self, **kwargs):
        if not kwargs.get('owner'):
            kwargs['owner'] = self.user

        resp = yield from self.model.add(self.user, **kwargs)

        json_resp = resp.to_json()
        return json_resp

    @gen.coroutine
    def get(self, *args):
        resp = yield from self.get_item(**self.params)
        json_resp = resp.to_json()
        return json_resp

    @gen.coroutine
    def post(self, *args):
        resp = yield from self.add(**self.params)
        self.write(resp)

    @gen.coroutine
    def delete(self, *args):
        item = yield from self.get_item(**self.params)
        resp = yield from item.delete()
        return resp


class RepositoryHandler(BaseModelHandler):

    @gen.coroutine
    def post(self, *args):
        if 'add-branch' in args:
            yield from self.add_branch()
            return

        elif 'remove-branch' in args:
            yield from self.remove_branch()
            return

        elif 'enable-plugin' in self.request.uri:
            yield from self.enable_plugin()
            return

        elif 'disable-plugin' in self.request.uri:
            yield from self.disable_plugin()
            return

        elif'start-build' not in args:
            yield super().post(*args)
            return

        ret = yield from self.start_build()
        self.write(ret)

    @asyncio.coroutine
    def enable_plugin(self):

        repo = yield from self.get_item(repo_name=self.params.get('name'))
        del self.params['name']
        plugin_name = self.params.get('plugin_name')
        del self.params['plugin_name']
        r = yield from repo.enable_plugin(plugin_name, **self.params)
        return r

    @asyncio.coroutine
    def disable_plugin(self):
        repo = yield from self.get_item(repo_name=self.params.get('name'))
        plugin_name = self.params.get('plugin_name')
        r = yield from repo.disable_plugin(name=plugin_name)
        return r

    @asyncio.coroutine
    def list_plugins(self):
        plugins = yield from Plugin.list()
        return plugins

    @asyncio.coroutine
    def start_build(self):
        item = yield from self.get_item(repo_name=self.params.get('name'))
        del self.params['name']
        ret = yield from item.start_build(**self.params)
        return ret

    @asyncio.coroutine
    def add_branch(self):
        item = yield from self.get_item(repo_name=self.params.get('name'))
        del self.params['name']
        notify = self.params['notify_only_latest']
        notify = True if notify == 'true' else False
        self.params['notify_only_latest'] = notify
        r = yield from item.add_branch(**self.params)
        return r

    @asyncio.coroutine
    def remove_branch(self):
        item = yield from self.get_item(repo_name=self.params.get('name'))
        del self.params['name']
        r = yield from item.remove_branch(**self.params)
        return r

    @gen.coroutine
    def prepare(self):
        super().prepare()
        if 'start-build' in self.request.uri:
            self._prepare_start_build()

        elif 'add-branch' in self.request.uri:
            kw = {'name': self.params.get('name'),
                  'branch_name': self.params.get('branch_name'),
                  'notify_only_latest': self.params.get('notify_only_latest')}
            self.params = kw

        elif 'remove-branch' in self.request.uri:
            kw = {'name': self.params.get('name'),
                  'branch_name': self.params.get('branch_name')}
            self.params = kw

        elif ('enable-plugin' in self.request.uri or
              'disable-plugin' in self.request.uri):
            yield from self._prepare_for_plugin()

        else:
            kw = {}
            kw['name'] = self.params.get('name')
            kw['url'] = self.params.get('url')
            kw['vcs_type'] = self.params.get('vcs_type')
            kw['update_seconds'] = self.params.get('update_seconds')
            kw['parallel_builds'] = self.params.get('parallel_builds')
            kw['slaves'] = self.params.getlist('slaves')
            self.params = kw

    @gen.coroutine
    def delete(self, *args):
        self.params = {'repo_name': self.params.get('name')}
        yield super().delete(*args)

    @gen.coroutine
    def put(self, *args):
        item = yield from self.get_item(repo_name=self.params['name'])
        del self.params['name']
        r = yield from item.update(**self.params)
        self.write(r)

    def _prepare_start_build(self):
        kw = {}
        kw['name'] = self.params.get('name')
        kw['builder_name'] = self.params.get('builder_name')
        kw['branch'] = self.params.get('branch')
        kw['slaves'] = self.params.getlist('slaves')
        kw['named_tree'] = self.params.get('named_tree')

        self.params = kw

    @asyncio.coroutine
    def _prepare_for_plugin(self):

        kw = {}
        plugin_name = self.params.get('plugin_name')
        plugin = yield from Plugin.get(self.user, name=plugin_name)
        for k, v in self.params.items():
            try:
                kw[k] = v[0] if getattr(plugin, k)['type'] != 'list' else [
                    i.strip() for i in v[0].split(',')]
            except (AttributeError, TypeError):
                # TypeError happens when a attribute is not a dict
                # ie, plugin pretty_name and description
                kw[k] = v[0]

        self.params = kw


class SlaveHandler(BaseModelHandler):

    def prepare(self):
        super().prepare()
        kw = {}
        kw['name'] = self.params.get('name')
        kw['host'] = self.params.get('host')
        kw['port'] = self.params.get('port')
        kw['token'] = self.params.get('token')
        self.params = kw

    @gen.coroutine
    def delete(self, *args):
        self.params = {'slave_name': self.params.get('name')}
        yield super().delete(*args)

    @gen.coroutine
    def put(self, *args):
        item = yield from self.get_item(slave_name=self.params['name'])
        yield from item.update(**self.params)


class StreamHandler(LoggerMixin, LoggedTemplateHandler, WebSocketHandler):

    def initialize(self):
        self.action = None
        self.repo_id = None
        self.events = {'repo_status_changed': self._send_repo_status_info,
                       'build_started': self._send_build_info,
                       'build_finished': self._send_build_info,
                       'build_added': self._send_build_info,
                       'step_started': self._send_build_info,
                       'step_finished': self._send_build_info,
                       'step_output_info': self._send_step_output_info}
        # maps actions to message (event) types
        self.action_messages = {'repo-status': ['repo_status_changed'],
                                'builds': ['build_started', 'build_finished',
                                           'build_added', 'step_started',
                                           'step_finished'],
                                'step-output': ['step_output_info']}

    def _bad_message_type_logger(self, message):
        msg = 'Bad. message type: {}'.format(message['event_type'])
        self.log(msg, level='warning')

    def _get_repo_id(self):
        repo_id = None
        keys = ['repo_id', 'repository_id']
        for key in keys:
            try:
                repo_id = self.request.arguments.get(key)[0].decode()
                break
            except TypeError:
                pass

        return repo_id

    def open(self, action):
        self.action = action
        self.repo_id = self._get_repo_id()
        f = ensure_future(StreamConnector.plug(
            self.user, self.repo_id, self.receiver))
        return f

    def receiver(self, sender, **message):
        message_type = message.get('event_type')
        msg = 'message arrived: {}'.format(message_type)
        self.log(msg, level='debug')
        if message_type not in self.action_messages.get(self.action, []):
            msg = 'leaving receiver'
            self.log(msg, level='debug')
            return
        outfn = self.events.get(message_type, self._bad_message_type_logger)
        try:
            outfn(message)
        except Exception:
            msg = traceback.format_exc()
            self.log(msg, level='error')

    def _send_step_output_info(self, info):
        """Sends information about step output to the ws client.

        :param info: Message sent by the master"""
        step_uuid = self.request.arguments.get('uuid')[0].decode()
        uuid = info.get('uuid')
        if step_uuid == uuid:
            self.write2sock(info)

    def _send_build_info(self, info):
        """Sends information about builds to the ws client.

        :param info: The message sent by the master"""

        self._format_info_dt(info)
        self.write2sock(info)

    def _format_info_dt(self, info):
        started = info.get('started')
        if started and is_datetime(started):
            info['started'] = format_datetime(string2datetime(started))

        finished = info.get('finished')
        if finished and is_datetime(finished):
            info['finished'] = format_datetime(string2datetime(finished))

        created = info.get('created')
        if created and is_datetime(created):
            info['created'] = format_datetime(string2datetime(created))

        buildset = info.get('buildset')
        if buildset:
            self._format_info_dt(buildset)

    def _send_repo_status_info(self, info):
        self.write2sock(info)

    def on_close(self):
        StreamConnector.unplug(self.user, self.repo_id, self.receiver)

    def write2sock(self, body):
        try:
            self.write_message(body)
        except WebSocketError:
            self.log('WebSocketError', level='debug')


class MainHandler(LoggedTemplateHandler):
    main_template = 'main.html'

    @gen.coroutine
    def get(self):
        repos = yield from Repository.list(self.user)
        slaves = yield from Slave.list(self.user)
        plugins = yield from Plugin.list(self.user)

        context = {'repos': repos, 'slaves': slaves,
                   'get_btn_class': self._get_btn_class,
                   'plugins': plugins}
        self.render_template(self.main_template, context)

    def _get_btn_class(self, status):
        return {'success': 'success', 'fail': 'danger',
                'running': 'info', 'exception': 'exception',
                'clone-exception': 'exception', 'ready': 'success',
                'warning': 'warning', 'cloning': 'pending'}.get(status)


class WaterfallHandler(LoggedTemplateHandler):
    template = 'waterfall.html'

    def prepare(self):
        super().prepare()
        self.params = PyroRequest(self.request.arguments)

    @gen.coroutine
    def get(self, repo_name):
        buildsets = yield from BuildSet.list(self.user, repo_name=repo_name)
        builders = yield from self._get_builders_for_buildsets(buildsets)
        repo = yield from Repository.get(self.user, repo_name=repo_name)

        def _ordered_builds(builds):
            return sorted(
                builds, key=lambda b: builders[builders.index(b.builder)].name)

        def fmtdt(dt):  # pragma: no cover
            # when the attribute is not set, it is a empty string int the
            # template, so we simply skip it here
            if not isinstance(dt, datetime.datetime):
                return
            return format_datetime(dt)

        context = {'buildsets': buildsets, 'builders': builders,
                   'ordered_builds': _ordered_builds,
                   'get_ending': self._get_ending,
                   'repository': repo, 'fmtdt': fmtdt}
        self.render_template(self.template, context)

    @asyncio.coroutine
    def _get_builders_for_buildsets(self, buildsets):
        builders = set()
        buildsets = buildsets or []
        for buildset in buildsets:
            for build in buildset.builds:
                builders.add(build.builder)

        # Now the thing here is: the builders here are made
        # from the response of buildset-list. It returns only
        # the builder id for builds, so now I retrieve the
        # 'full' builder using builder-list
        ids = [b.id for b in builders]
        builders = yield from Builder.list(self.user, id__in=ids)
        builders_dict = {b.id: b for b in builders}
        for buildset in buildsets:
            for build in buildset.builds:
                build.builder = builders_dict[build.builder.id]

        return sorted(builders, key=lambda b: b.name)

    def _get_ending(self, build, build_index, builders):
        i = build_index
        while build.builder != builders[i] and len(builders) > i:
            tag = '</td><td class="builder-column builder-column-id-{}'
            tag += 'builder-column-index-{}">'
            yield tag.format(builders[i].id, i + 1)
            i += 1
        yield ''


url = URLSpec('/$', MainHandler)
waterfall = URLSpec('/waterfall/(.*)', WaterfallHandler)
websocket = URLSpec('/api/socks/(.*)', StreamHandler)
login = URLSpec('/(login|logout)', LoginHandler)

app = PyroApplication([url, waterfall, login, websocket])

static_app = StaticApplication()

repo_kwargs = {'model': Repository}
repo_api_url = URLSpec('/api/repo/(.*)', RepositoryHandler, repo_kwargs)

slave_kwargs = {'model': Slave}
slave_api_url = URLSpec('/api/slave/(.*)', SlaveHandler, slave_kwargs)

api_app = PyroApplication([repo_api_url, slave_api_url])



# -*- coding: utf-8 -*-

# Copyright 2016 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from toxicbuild.core.exceptions import PluginNotFound


class PluginMeta(type):

    def __new__(cls, name, bases, attrs):
        no_list = attrs.get('no_list')
        r = super().__new__(cls, name, bases, attrs)
        if no_list is None:
            setattr(r, 'no_list', False)
        return r


class Plugin(metaclass=PluginMeta):
    """This is a base plugin. Plugins may implement aditional behavior
    to your builds."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # Your plugin must have an unique name
    name = 'BaseCorePlugin'

    @classmethod
    def list_plugins(cls, plugin_type=None):
        """Returns a list of Plugin subclasses.

        :param plugin_type: the plugin's type."""
        plugins = []
        for plugin in cls.__subclasses__():
            plugins += plugin.list_plugins(plugin_type=plugin_type)

        return plugins + [p for p in cls.__subclasses__() if not p.no_list]

    @classmethod
    def get_plugin(cls, name):
        """ Returns a Plugin subclass based on its name.

        :param name: Plugin's name."""

        for plugin in cls.__subclasses__():
            try:
                return plugin.get_plugin(name)
            except PluginNotFound:
                pass

            if plugin.name == name and not plugin.no_list:
                return plugin

        raise PluginNotFound('Plugin {} does not exist.'.format(name))



# -*- coding: utf-8 -*-

# Copyright 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.
from email.mime.text import MIMEText
import aiosmtplib
from toxicbuild.core.utils import LoggerMixin
from toxicbuild.master import settings


class MailSenderNotConnected(Exception):
    pass


class MailSender(LoggerMixin):
    """Simple mail sender. Takes host/port/auth params from
    settings.

    To send an email, use the context manager:

    .. code-block:: python

       recipients = ['me@mail.com', 'other@mail.com']
       async with MailSender(recipients) as sender:
           await sender.send('Subject', 'This is the message body')

    """

    def __init__(self, recipients):
        self.recipients = recipients
        self.smtp = None
        self.mail_from = settings.SMTP_MAIL_FROM
        self.host = settings.SMTP_HOST
        self.port = settings.SMTP_PORT
        self.username = settings.SMTP_USERNAME
        self.passwd = settings.SMTP_PASSWORD
        self.validate_certs = settings.SMTP_VALIDATE_CERTS
        self.starttls = settings.SMTP_STARTTLS
        self._connected = False

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc, value, traceback):
        await self.disconnect()

    async def connect(self):
        """Connects to a smtp server."""

        if self._connected:
            msg = 'Already connected to smtp server. Leaving...'
            self.log(msg, level='warning')
            return

        self.smtp = aiosmtplib.SMTP(hostname=self.host, port=self.port,
                                    validate_certs=self.validate_certs)
        if self.starttls:
            await self.smtp.connect(use_tls=False)
            await self.smtp.starttls()
        else:
            await self.smtp.connect()

        await self.smtp.login(self.username, self.passwd)
        self._connected = True

    async def disconnect(self):
        """Closes the connection to the smpt server"""

        await self.smtp.quit()
        self.smtp.close()
        self.smtp = None
        self._connected = False

    async def send(self, subject, message):
        """Send an email message.

        :param subject: Email subject.
        :param message: Email body
        """

        if not self._connected:
            msg = 'You must connect before sending emails'
            raise MailSenderNotConnected(msg)

        msg = MIMEText(message)
        msg['From'] = self.mail_from
        msg['subject'] = subject
        await self.smtp.send_message(msg, recipients=self.recipients)



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

# Welcome to the strange land of the user interface hole,
# the place where master's clients can ask for what they need...
# ... e onde tudo pode acontecer...
# In fact, boring module!

import asyncio
from asyncio import ensure_future
import inspect
import json
import traceback
from toxicbuild.core import BaseToxicProtocol
from toxicbuild.core.utils import LoggerMixin
from toxicbuild.master import settings
from toxicbuild.master.build import BuildSet, Builder
from toxicbuild.master.repository import Repository, RepositoryRevision
from toxicbuild.master.exceptions import (UIFunctionNotFound,
                                          OwnerDoesNotExist, NotEnoughPerms)
from toxicbuild.master.plugins import MasterPlugin
from toxicbuild.master.slave import Slave
from toxicbuild.master.signals import (step_started, step_finished,
                                       build_started, build_finished,
                                       repo_status_changed, build_added,
                                       step_output_arrived, repo_added)
from toxicbuild.master.users import User, Organization


class UIHole(BaseToxicProtocol, LoggerMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None

    salt = settings.BCRYPT_SALT
    encrypted_token = settings.ACCESS_TOKEN

    async def client_connected(self):

        data = self.data.get('body') or {}

        if self.action != 'user-authenticate':
            # when we are authenticating we don't need (and we can't have)
            # a requester user, so we only try to get it when we are not
            # authenticating.
            try:
                user_id = self.data.get('user_id', '')
                if not user_id:
                    raise User.DoesNotExist
                self.user = await User.objects.get(id=user_id)
            except User.DoesNotExist:
                msg = 'User {} does not exist'.format(user_id)
                self.log(msg, level='warning')
                status = 2
                await self.send_response(code=status, body={'error': msg})
                self.close_connection()
                return status

        if self.action == 'stream':
            handler = UIStreamHandler(self)
        else:
            handler = HoleHandler(data, self.action, self)

        try:
            await handler.handle()
            status = 0

        except NotEnoughPerms:
            msg = 'User {} does not have enough permissions.'.format(
                str(self.user.id))
            self.log(msg, level='warning')
            status = 3
            await self.send_response(code=status, body={'error': msg})
        except Exception:
            msg = traceback.format_exc()
            status = 1
            await self.send_response(code=1, body={'error': msg})
            self.close_connection()

        return status


class HoleHandler:

    """ Handles the incomming connections for the UIHole. It has the following
    methods available to the clients:

    * `repo-add`
    * `repo-get`
    * `repo-list`
    * `repo-remove`
    * `repo-update`
    * `repo-add-slave`
    * `repo-remove-slave`
    * `repo-add-branch`
    * `repo-remove-branch`
    * `repo-enable-plugin`
    * `repo-disable-plugin`
    * `repo-start-build`
    * `slave-add`
    * `slave-get`
    * `slave-list`
    * `slave-remove`
    * `slave-update`
    * `plugins-list`
    * `plugin-get`
    * `buildset-list`
    * `builder-show`
    * `list-funcs`
    * `user-add`
    * `user-remove`
    * `user-authenticate`
    """

    def __init__(self, data, action, protocol):
        self.data = data
        self.action = action
        self.protocol = protocol

    async def handle(self):

        attrname = self.action.replace('-', '_')
        if attrname not in self._get_action_methods():
            raise UIFunctionNotFound(self.action)

        func = getattr(self, attrname)
        r = func(**self.data)
        if asyncio.coroutines.iscoroutine(r):
            r = await r

        await self.protocol.send_response(code=0, body=r)
        self.protocol.close_connection()

    def _get_method_signature(self, method):
        sig = inspect.signature(method)
        doc = method.__doc__ or ''
        siginfo = {'doc': doc, 'parameters': []}

        for name, param in sig.parameters.items():
            pinfo = {'name': name}
            required = '=' not in str(param)
            pinfo['required'] = required
            if not required:
                default = param.default
                pinfo['default'] = default

            siginfo['parameters'].append(pinfo)

        return siginfo

    def _user_is_allowed(self, what):
        if self.protocol.user.is_superuser or \
           what in self.protocol.user.allowed_actions:
            return True
        return False

    async def user_add(self, email, password, allowed_actions, username=None):
        """Adds a new user.

        :param email: User email.
        :param password: User password
        :param allowed_actions: What the user can do.
        :param username: Username for the user."""

        if not self._user_is_allowed('add_user'):
            raise NotEnoughPerms

        user = User(email=email, allowed_actions=allowed_actions,
                    username=username)
        user.set_password(password)
        await user.save()
        return {'user-add': {'id': str(user.id), 'username': user.username,
                             'email': user.email}}

    async def user_remove(self, **kwargs):
        """Removes a user from the system."""

        if not self._user_is_allowed('remove_user'):
            raise NotEnoughPerms

        user = await User.objects.get(**kwargs)
        await user.delete()
        return {'user-remove': 'ok'}

    async def user_authenticate(self, username_or_email, password):
        """Authenticates an user. Returns user.to_dict() is
        authenticated. Raises ``InvalidCredentials`` if a user with
        this credentials does not exist.

        :param username_or_email: Username or email to use to authenticate.
        :param password: Not encrypted password."""

        user = await User.authenticate(username_or_email, password)
        return {'user-authenticate': user.to_dict()}

    async def repo_add(self, repo_name, repo_url, owner_id,
                       update_seconds, vcs_type, slaves=None,
                       parallel_builds=None):
        """ Adds a new repository and first_run() it.

        :param repo_name: Repository name
        :param repo_url: Repository vcs url
        :param owner_id: Id of the repository's owner.
        :param update_seconds: Time to poll for changes
        :param vcs_type: Type of vcs being used.
        :param slaves: A list of slave names.
        :params parallel_builds: How many parallel builds this repository
          executes. If None, there is no limit."""

        if not self._user_is_allowed('add_repo'):
            raise NotEnoughPerms

        repo_name = repo_name.strip()
        repo_url = repo_url.strip()
        slaves_info = slaves or []
        slaves = []
        for name in slaves_info:
            slave = await Slave.get(name=name)
            slaves.append(slave)

        kw = {}
        if parallel_builds:
            kw['parallel_builds'] = parallel_builds

        owner = await self._get_owner(owner_id)

        repo = await Repository.create(
            repo_name, repo_url, owner, update_seconds, vcs_type,
            slaves, **kw)
        repo_dict = await self._get_repo_dict(repo)
        return {'repo-add': repo_dict}

    async def _get_owner(self, owner_id):
        owner_types = [User, Organization]
        for owner_type in owner_types:
            try:
                owner = await owner_type.objects.get(id=owner_id)
                return owner
            except owner_type.DoesNotExist:
                pass

        msg = 'The owner {} does not exist'.format(owner_id)
        raise OwnerDoesNotExist(msg)

    async def repo_get(self, repo_name=None, repo_url=None):
        """Shows information about one specific repository. One of
        ``repo_name`` or ``repo_url`` is required.

        :param repo_name: Repository name,
        :param repo_url: Repository vcs url."""

        if not (repo_name or repo_url):
            raise TypeError("repo_name or repo_url required")

        kw = {}
        if repo_name:
            kw['name'] = repo_name

        if repo_url:
            kw['url'] = repo_url

        repo = await Repository.get_for_user(self.protocol.user, **kw)
        repo_dict = await self._get_repo_dict(repo)
        return {'repo-get': repo_dict}

    async def repo_remove(self, repo_name):
        """ Removes a repository from toxicubild.

        :param repo_name: Repository name."""

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        await repo.remove()
        return {'repo-remove': 'ok'}

    async def repo_list(self):
        """ Lists all repositories. """

        repos = Repository.list_for_user(self.protocol.user)
        repo_list = []
        async for repo in repos:

            repo_dict = await self._get_repo_dict(repo)
            repo_list.append(repo_dict)

        return {'repo-list': repo_list}

    async def repo_update(self, repo_name, **kwargs):
        """ Updates repository information.

        :param repo_name: Repository name
        :param kwargs: kwargs to update the repository"""

        if kwargs.get('slaves'):
            qs = Slave.objects(name__in=kwargs.get('slaves'))
            slaves_instances = await qs.to_list()
            kwargs['slaves'] = slaves_instances

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        [setattr(repo, k, v) for k, v in kwargs.items()]

        await repo.save()
        return {'repo-update': 'ok'}

    async def repo_add_slave(self, repo_name, slave_name):
        """ Adds a slave to a repository.

        :param repo_name: Repository name.
        :param slave_name: Slave name."""

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        slave = await Slave.get_for_user(self.protocol.user,
                                         name=slave_name)
        await repo.add_slave(slave)
        return {'repo-add-slave': 'ok'}

    async def repo_remove_slave(self, repo_name, slave_name):
        """ Removes a slave from toxicbuild. """

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)

        slave = await Slave.get_for_user(self.protocol.user,
                                         name=slave_name)
        await repo.remove_slave(slave)
        return {'repo-remove-slave': 'ok'}

    async def repo_add_branch(self, repo_name, branch_name,
                              notify_only_latest=False):
        """Adds a branch to the list of branches of the repository.

        :param repo_name: Reporitory name
        :param branch_name: Branch's name
        :notify_only_latest: If True only the latest commit in the
          branch will trigger a build."""
        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        await repo.add_or_update_branch(branch_name, notify_only_latest)
        return {'repo-add-branch': 'ok'}

    async def repo_remove_branch(self, repo_name, branch_name):
        """Removes a branch from the list of branches of a repository.
        :param repo_name: Repository name
        :param branch_name: Branch's name."""

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        await repo.remove_branch(branch_name)
        return {'repo-remove-branch': 'ok'}

    async def repo_enable_plugin(self, repo_name, plugin_name, **kwargs):
        """Enables a plugin to a repository.

        :param repo_name: Repository name.
        :param plugin_name: Plugin name
        :param kwargs: kwargs passed to the plugin."""

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        await repo.enable_plugin(plugin_name, **kwargs)
        return {'repo-enable-plugin': 'ok'}

    async def repo_disable_plugin(self, repo_name, **kwargs):
        """Disables a plugin from a repository.

        :param repo_name: Repository name.
        :param kwargs: kwargs passed to the plugin"""

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        await repo.disable_plugin(**kwargs)
        return {'repo-disable-plugin': 'ok'}

    async def repo_start_build(self, repo_name, branch, builder_name=None,
                               named_tree=None, slaves=[]):
        """ Starts a(some) build(s) in a given repository. """
        # Mutable stuff on method declaration. Sin!!! Take that, PyLint!

        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)

        slave_instances = []
        for sname in slaves:
            slave = await Slave.get(name=sname)
            slave_instances.append(slave)

        slaves = slave_instances

        if not slaves:
            slaves = await repo.slaves

        if not named_tree:
            rev = await repo.get_latest_revision_for_branch(branch)
            named_tree = rev.commit
        else:
            rev = await RepositoryRevision.get(repository=repo,
                                               branch=branch,
                                               commit=named_tree)

        if not builder_name:
            builders = await self._get_builders(slaves, rev)
        else:
            blist = [(await Builder.get(name=builder_name,
                                        repository=repo))]
            builders = {}
            for slave in slaves:
                builders.update({slave: blist})

        builds_count = 0

        buildset = await BuildSet.create(repository=repo, revision=rev,
                                         save=False)
        for slave in slaves:
            await repo.add_builds_for_slave(buildset, slave,
                                            builders[slave])

        return {'repo-start-build': '{} builds added'.format(builds_count)}

    async def slave_add(self, slave_name, slave_host, slave_port, slave_token,
                        owner_id):
        """ Adds a new slave to toxicbuild.

        :param slave_name: A name for the slave,
        :param slave_host: Host where the slave is.
        :param slave_port: Port to connect to the slave
        :param slave_token: Auth token for the slave.
        :param owner_id: Slave's owner id."""

        if not self._user_is_allowed('add_slave'):
            raise NotEnoughPerms

        owner = await self._get_owner(owner_id)
        slave = await Slave.create(name=slave_name, host=slave_host,
                                   port=slave_port, token=slave_token,
                                   owner=owner)

        slave_dict = self._get_slave_dict(slave)
        return {'slave-add': slave_dict}

    async def slave_get(self, slave_name):
        """Returns information about one specific slave"""

        slave = await Slave.get_for_user(self.protocol.user,
                                         name=slave_name)
        slave_dict = self._get_slave_dict(slave)
        return {'slave-get': slave_dict}

    async def slave_remove(self, slave_name):
        """ Removes a slave from toxicbuild. """

        slave = await Slave.get_for_user(self.protocol.user,
                                         name=slave_name)

        await slave.delete()

        return {'slave-remove': 'ok'}

    async def slave_list(self):
        """ Lists all slaves. """

        slaves = Slave.list_for_user(self.protocol.user)
        slave_list = []

        async for slave in slaves:
            slave_dict = self._get_slave_dict(slave)
            slave_list.append(slave_dict)

        return {'slave-list': slave_list}

    async def slave_update(self, slave_name, **kwargs):
        """Updates infomation of a slave."""

        slave = await Slave.get_for_user(self.protocol.user, name=slave_name)
        [setattr(slave, k, v) for k, v in kwargs.items()]

        await slave.save()
        return {'slave-update': 'ok'}

    async def buildset_list(self, repo_name=None, skip=0, offset=None):
        """ Lists all buildsets.

        If ``repo_name``, only builders from this repository will be listed.
        :param repo_name: Repository's name.
        :param skip: skip for buildset list.
        :param offset: offset for buildset list.
        """

        buildsets = BuildSet.objects.no_dereference()
        if repo_name:
            repository = await Repository.get_for_user(
                self.protocol.user, name=repo_name)
            buildsets = buildsets.filter(repository=repository)

        buildsets = buildsets.order_by('-created')
        count = await buildsets.count()

        stop = count if not offset else skip + offset

        buildsets = buildsets[skip:stop]
        buildset_list = []
        buildsets = await buildsets.to_list()
        for b in buildsets:
            bdict = b.to_dict(id_as_str=True)
            buildset_list.append(bdict)

        return {'buildset-list': buildset_list}

    async def builder_list(self, **kwargs):
        """List builders.

        :param kwargs: Arguments to filter the list."""

        queryset = Builder.objects.filter(**kwargs)
        builders = await queryset.to_list()
        blist = []

        for b in builders:
            blist.append((await b.to_dict(id_as_str=True)))

        return {'builder-list': blist}

    def plugins_list(self):
        """Lists all plugins available to the master."""

        plugins = MasterPlugin.list_plugins()
        plugins_schemas = [p.get_schema(to_serialize=True) for p in plugins]
        return {'plugins-list': plugins_schemas}

    def plugin_get(self, **kwargs):
        """Returns a specific plugin."""

        name = kwargs.get('name')
        plugin = MasterPlugin.get_plugin(name=name)
        return {'plugin-get': plugin.get_schema(to_serialize=True)}

    async def builder_show(self, repo_name, builder_name, skip=0, offset=None):
        """ Returns information about one specific builder.

        :param repo_name: The builder's repository name.
        :param builder_name: The bulider's name.
        :param skip: How many elements we should skip in the result.
        :param offset: How many results we should return.
        """

        kwargs = {'name': builder_name}
        repo = await Repository.get_for_user(self.protocol.user,
                                             name=repo_name)
        kwargs.update({'repository': repo})

        builder = await Builder.get(**kwargs)
        buildsets = BuildSet.objects(builds__builder=builder)
        count = await buildsets.count()
        stop = count if not offset else skip + offset
        buildsets = buildsets[skip:stop]
        buildsets = await buildsets.to_list()
        buildsets_list = []
        for buildset in buildsets:
            bdict = buildset.to_dict()
            bdict['builds'] = []
            for b in (await buildset.get_builds_for(builder=builder)):
                build_dict = b.to_dict()
                bdict['builds'].append(build_dict)

            buildsets_list.append(bdict)

        builder_dict = await builder.to_dict()
        builder_dict['buildsets'] = buildsets_list
        return {'builder-show': builder_dict}

    def list_funcs(self):
        """ Lists the functions available for user interfaces. """

        funcs = self._get_action_methods()

        funcs = {n.replace('_', '-'): self._get_method_signature(m)
                 for n, m in funcs.items()}

        return {'list-funcs': funcs}

    def _get_action_methods(self):
        """ Returns the methods that are avaliable as actions for users. """
        forbiden = ['handle', 'protocol', 'log']

        func_names = [n for n in dir(self) if not n.startswith('_') and
                      n not in forbiden and callable(getattr(self, n))]

        funcs = {n: getattr(self, n) for n in func_names}
        return funcs

    async def _get_repo_dict(self, repo):
        """Returns a dictionary for a given repository"""

        repo_dict = json.loads(repo.to_json())
        repo_dict['id'] = str(repo.id)
        repo_dict['status'] = await repo.get_status()
        slaves = await repo.slaves
        repo_dict['slaves'] = [self._get_slave_dict(s) for s in slaves]
        repo_dict['parallel_builds'] = repo.parallel_builds or ''
        for p in repo_dict['plugins']:
            p['name'] = p['_name']

        return repo_dict

    def _get_slave_dict(self, slave):
        slave_dict = json.loads(slave.to_json())
        slave_dict['id'] = str(slave.id)
        return slave_dict

    async def _get_builders(self, slaves, revision):
        repo = await revision.repository
        builders = {}
        for slave in slaves:
            builders[slave] = await repo.build_manager.get_builders(
                slave, revision)

        return builders


class UIStreamHandler(LoggerMixin):

    """ Handler that keeps the connection open and messages when
    builds and steps are stated or finished.
    """

    def __init__(self, protocol):
        self.protocol = protocol

        def connection_lost_cb(exc):  # pragma no cover
            self._disconnectfromsignals()

        self.protocol.connection_lost_cb = connection_lost_cb

    async def step_started(self, sender, **kw):
        await self.send_info('step_started', **kw)

    async def step_finished(self, sender, **kw):
        await self.send_info('step_finished', **kw)

    async def build_started(self, sender, **kw):
        await self.send_info('build_started', **kw)

    async def build_finished(self, sender, **kw):
        await self.send_info('build_finished', **kw)

    async def build_added(self, sender, **kw):
        await self.send_info('build_added', **kw)

    async def _connect2signals(self):
        repos = Repository.list_for_user(self.protocol.user)
        async for repo in repos:
            self._connect_repo(repo)
        repo_added.connect(self.check_repo_added)

    def _connect_repo(self, repo):
        step_started.connect(self.step_started, sender=str(repo.id))
        step_finished.connect(self.step_finished, sender=str(repo.id))
        build_started.connect(self.build_started, sender=str(repo.id))
        build_finished.connect(self.build_finished, sender=str(repo.id))
        repo_status_changed.connect(self.send_repo_status_info,
                                    sender=str(repo.id))
        build_added.connect(self.build_added, sender=str(repo.id))
        step_output_arrived.connect(self.send_step_output_info,
                                    sender=str(repo.id))

    async def check_repo_added(self, sender, **kw):
        try:
            repo = await Repository.get_for_user(self.protocol.user,
                                                 id=sender)
        except NotEnoughPerms:
            return

        self._connect_repo(repo)

    def _disconnectfromsignals(self):
        step_output_arrived.disconnect(self.send_step_output_info)
        step_started.disconnect(self.step_started)
        step_finished.disconnect(self.step_finished)
        build_started.disconnect(self.build_started)
        build_finished.disconnect(self.build_finished)
        repo_status_changed.disconnect(self.send_repo_status_info)
        build_added.disconnect(self.build_added)
        repo_added.disconnect(self.check_repo_added)

    async def handle(self):
        await self._connect2signals()
        await self.protocol.send_response(code=0, body={'stream': 'ok'})

    async def send_info(self, info_type, build=None, step=None):
        repo = await build.repository
        slave = await build.slave

        build_dict = build.to_dict(id_as_str=True)
        slave = slave.to_dict(id_as_str=True)
        repo = await repo.to_dict(id_as_str=True)
        buildset = await build.get_buildset()

        build_dict['slave'] = slave
        build_dict['repository'] = repo
        build_dict['buildset'] = buildset.to_dict(id_as_str=True)

        if step:
            step = step.to_dict()
            step['build'] = build_dict
            info = step
        else:
            info = build_dict

        final_info = {'event_type': info_type}
        final_info.update(info)

        f = ensure_future(self.send_response(code=0, body=final_info))

        return f

    async def send_repo_status_info(self, repo_id, old_status, new_status):
        """Called by the signal ``repo_status_changed``

        :param repo_id: Id of the repository that had its status changed.
        :param old_status: The old status of the repository
        :param new_status: The new status of the repostiory."""

        repo = await Repository.get(id=repo_id)
        rdict = await repo.to_dict(id_as_str=True)
        rdict['status'] = new_status
        rdict['old_status'] = old_status
        rdict['event_type'] = 'repo_status_changed'
        f = ensure_future(self.send_response(code=0, body=rdict))
        return f

    def send_step_output_info(self, repo, step_info):
        """Called by the signal ``step_output_arrived``.

        :param repo: The repository that is building something.
        :param step_info: The information about the step output."""

        step_info['event_type'] = 'step_output_info'
        f = ensure_future(self.send_response(code=0, body=step_info))
        return f

    async def send_response(self, code, body):
        try:
            await self.protocol.send_response(code=code, body=body)
        except ConnectionResetError:
            self.protocol._transport.close()
            self._disconnectfromsignals()


class HoleServer:

    def __init__(self, addr='127.0.0.1', port=6666):
        self.protocol = UIHole
        self.loop = asyncio.get_event_loop()
        self.addr = addr
        self.port = port

    def serve(self):

        coro = self.loop.create_server(
            self.get_protocol_instance, self.addr,  self.port)

        ensure_future(coro)

    def get_protocol_instance(self):
        return self.protocol(self.loop)



# -*- coding: utf-8 -*-
"""This module implements a simple asynchronous interface for
http requests.

Usage:
------

.. code-block:: python

    from toxicbuild.core import requests
    response = yield from requests.get('http://google.com/')
    print(response.text)
"""

# Copyright 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import aiohttp


class Response:
    """Encapsulates a response from a http request"""

    def __init__(self, status, text):
        """Constructor for Response.

        :param status: The response status.
        :param text: The response text."""
        self.status = status
        self.text = text


@asyncio.coroutine
def _request(method, url, **kwargs):
    """Performs a http request and returns an instance of
    :class:`toxicbuild.core.requests.Response`

    :param method: The requrest's method.
    :param url: Request's url.
    :param kwargs: Arguments passed to aiohttp.ClientSession.request
        method.
    """

    loop = asyncio.get_event_loop()

    client = aiohttp.ClientSession(loop=loop)
    try:
        resp = yield from client.request(method, url, **kwargs)
        status = resp.status
        text = yield from resp.text()
        yield from resp.release()
    finally:
        client.close()

    return Response(status, text)


@asyncio.coroutine
def get(url, **kwargs):
    """Performs a http GET request

    :param url: Request's url.
    :param kwargs: Args passed to :func:`toxicbuild.core.requests._request`.
    """

    method = 'GET'
    resp = yield from _request(method, url, **kwargs)
    return resp


@asyncio.coroutine
def post(url, **kwargs):
    """Performs a http POST request

    :param url: Request's url.
    :param kwargs: Args passed to :func:`toxicbuild.core.requests._request`.
    """

    method = 'POST'
    resp = yield from _request(method, url, **kwargs)
    return resp


@asyncio.coroutine
def put(url, **kwargs):
    """Performs a http PUT request

    :param url: Request's url.
    :param kwargs: Args passed to :func:`toxicbuild.core.requests._request`.
    """

    method = 'PUT'
    resp = yield from _request(method, url, **kwargs)
    return resp


@asyncio.coroutine
def delete(url, **kwargs):
    """Performs a http DELETE request

    :param url: Request's url.
    :param kwargs: Args passed to :func:`toxicbuild.core.requests._request`.
    """

    method = 'DELETE'
    resp = yield from _request(method, url, **kwargs)
    return resp



# -*- coding: utf-8 -*-

# Copyright 2015, 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import traceback
from toxicbuild.core.protocol import BaseToxicProtocol
from toxicbuild.core.utils import log, datetime2string, now
from toxicbuild.slave import settings
from toxicbuild.slave.managers import BuildManager
from toxicbuild.slave.exceptions import (BadData, BadBuilderConfig)


class BuildServerProtocol(BaseToxicProtocol):

    """ A simple server for build requests.
    """
    salt = settings.BCRYPT_SALT
    encrypted_token = settings.ACCESS_TOKEN

    @asyncio.coroutine
    def client_connected(self):
        try:
            self.log('executing {} for {}'.format(self.action, self.peername))
            status = 0
            if self.action == 'healthcheck':
                yield from self.healthcheck()

            elif self.action == 'list_builders':
                yield from self.list_builders()

            elif self.action == 'build':
                # build has a strange behavior. It sends messages to the client
                # directly. I think it should be an iterable here and the
                # protocol should send the messages. But how do to that?
                # I tought, instead of the iterable, use messages sent from
                # BuildManager and captured by the protocol, but didn't try
                # that yet. Any ideas?

                build_info = yield from self.build()

                yield from self.send_response(code=0, body=build_info)
            else:
                msg = 'Action {} does not exist'.format(self.action)
                self.log(msg, level='error')
                yield from self.send_response(code=1, body={'error': msg})

        except BadData:
            msg = 'Something wrong with your data {}'.format(self.raw_data)
            self.log('bad data', level='error')
            yield from self.send_response(code=1, body={'error': msg})
            status = 1
        except Exception as e:
            self.log(e.args[0], level='error')
            msg = traceback.format_exc()
            status = 1
            yield from self.send_response(code=1, body={'error': msg})

        finally:
            self.close_connection()

        return status

    @asyncio.coroutine
    def healthcheck(self):
        """ Informs that the server is up and running
        """
        yield from self.send_response(code=0, body='I\'m alive!')

    @asyncio.coroutine
    def list_builders(self):
        """ Informs all builders' names for this repo/branch/named_tree
        """
        with (yield from self.get_buildmanager()) as manager:
            # We do not work after wait because if we wait for it
            # the other instance working is in the same named_tree
            yield from manager.update_and_checkout(work_after_wait=False)

            builder_names = manager.list_builders()
        yield from self.send_response(code=0, body={'builders': builder_names})

    @asyncio.coroutine
    def build(self):
        """ Performs a build requested by the client using the params sent
        in the request data
        """
        with (yield from self.get_buildmanager()) as manager:
            yield from manager.update_and_checkout(work_after_wait=False)

            try:
                builder_name = self.data['body']['builder_name']
            except KeyError:
                raise BadData("No builder name for build.")

            try:
                builder = yield from manager.load_builder(builder_name)
            except BadBuilderConfig:
                build_info = {'steps': [], 'status': 'exception',
                              'started': datetime2string(now()),
                              'finished': datetime2string(now()),
                              'branch': manager.branch,
                              'named_tree': manager.named_tree}
            else:
                build_info = yield from builder.build()
            return build_info

    @asyncio.coroutine
    def get_buildmanager(self):
        """ Returns the builder manager for this request
        """
        try:
            repo_url = self.data['body']['repo_url']
            branch = self.data['body']['branch']
            vcs_type = self.data['body']['vcs_type']
            named_tree = self.data['body']['named_tree']
        except KeyError:
            raise BadData('Bad data!')

        manager = BuildManager(self, repo_url, vcs_type, branch, named_tree)

        return manager

    def log(self, msg, level='info'):
        log('[{}] {} '.format(type(self).__name__, msg), level)



# -*- coding: utf-8 -*-

# Copyright 2016 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

import asyncio
import os
import subprocess
import sys
from toxicbuild.core.cmd import command, main
from toxicbuild.core.utils import changedir
from toxicbuild.master import create as create_master
from toxicbuild.master import create_user
from toxicbuild.master import create_settings_and_connect
from toxicbuild.master.slave import Slave
from toxicbuild.slave import create as create_slave
from toxicbuild.ui import create as create_ui

main._generate_queue = []


@command
def create(root_dir):  # pragma no cover
    """ Creates a new toxicbuild environment.

    Environments for master, slave and ui will be created.

    :param --root_dir: Root directory for toxicbuild.
    """

    slave_root = os.path.join(root_dir, 'slave')
    master_root = os.path.join(root_dir, 'master')
    ui_root = os.path.join(root_dir, 'ui')

    # first we create a slave and a master
    slave_token = create_slave(slave_root)
    master_token = create_master(master_root)
    # a super user to access stuff
    conffile = os.path.join(master_root, 'toxicmaster.conf')
    create_user(conffile, superuser=True)

    with changedir(master_root):
        create_settings_and_connect()
        loop = asyncio.get_event_loop()
        # now we add this slave to the master
        slave = Slave(name='LocalSlave', token=slave_token,
                      host='localhost', port=7777)

        loop.run_until_complete(slave.save())

    # and finally create a web ui
    create_ui(ui_root, master_token)


@command
def start(workdir, loglevel='error'):  # pragma no cover
    """Starts toxicbuild.

    New instances for master, slave and web ui wil be started.

    :param workdir: Work directory for toxicbuild.
    :param --loglevel: Level for logging messages."""

    _call_processes(workdir, loglevel)


@command
def stop(workdir):  # pragma no cover
    """Stops toxicbuild

    The master, slave and web ui instances will be stopped.
    :param workdir: workdir for the instance to be stopped."""

    _call_processes(workdir, daemonize=False)


@command
def restart(workdir):
    """Restarts toxicbuild

    The instances of master, slave and web ui in ``workdir`` will be restarted.

    :param workdir: Workdir for master to be killed.
    """

    stop(workdir)
    start(workdir)


def _call_processes(workdir, loglevel=None, daemonize=True):  # pragma no cover

    # just pretend you didnt' see this function

    slave_root = os.path.join(workdir, 'slave')
    master_root = os.path.join(workdir, 'master')
    ui_root = os.path.join(workdir, 'ui')

    cmd = sys.argv[0].replace('-script', '')
    slave_cmd = cmd.replace('build', 'slave')
    master_cmd = cmd.replace('build', 'master')
    web_cmd = cmd.replace('build', 'web')

    slave_cmd_line = sys.argv[:]
    slave_cmd_line[0] = slave_cmd
    slave_cmd_line[2] = slave_root
    if daemonize:
        slave_cmd_line.append('--daemonize')

    if loglevel:
        slave_cmd_line.append('--loglevel')
        slave_cmd_line.append(loglevel)

    master_cmd_line = sys.argv[:]
    master_cmd_line[0] = master_cmd
    master_cmd_line[2] = master_root
    if daemonize:
        master_cmd_line.append('--daemonize')
    if loglevel:
        master_cmd_line.append('--loglevel')
        master_cmd_line.append(loglevel)

    web_cmd_line = sys.argv[:]
    web_cmd_line[0] = web_cmd
    web_cmd_line[2] = ui_root
    if daemonize:
        web_cmd_line.append('--daemonize')
    if loglevel:
        web_cmd_line.append('--loglevel')
        web_cmd_line.append(loglevel)

    subprocess.call(slave_cmd_line)
    subprocess.call(master_cmd_line)
    subprocess.call(web_cmd_line)


if __name__ == '__main__':  # pragma no cover
    main()



# -*- coding: utf-8 -*-
"""This module implements plugins meant to be used in reaction
to some signal sent by the master in the build process.

To implement your own plugins you must to subclass
:class:`toxicbuild.master.plugins.MasterPlugin` and implement a run() method.

The class :class:`toxicbuild.master.plugins.MasterPlugin` is a mongomotor's
document that you can subclass and create your own fields to store the
plugin's config params. It already has the following fields:

- repository: A reference field to a
  :class:`toxicbuild.master.repository.Repository`
- branches: A list of branch names that triggers the plugin.
- statuses: A list of statuses that triggers the plugin.

Example:
^^^^^^^^

.. code-block:: python

    from mongomotor.fields import StringField
    from toxicbuild.master.plugins import MasterPlugin

    class MyPlugin(MasterPlugin):

        # you must define name and type
        name = 'my-plugin'
        type = 'notification'
        # optionally you may define pretty_name and description
        pretty_name = "My Plugin"
        description = "A very cool plugin"
        something_to_store_on_database = PrettyStringField()

        async def run(self, sender):
            '''Here is where you implement your stuff. Sender is a
               repository instance.'''

"""

# Copyright 2016, 2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.

from asyncio import ensure_future
from collections import OrderedDict
import copy
import json
from uuid import uuid4
from mongomotor import EmbeddedDocument
from mongoengine.base.metaclasses import DocumentMetaclass
from mongomotor.fields import (StringField, URLField, ListField, UUIDField,
                               ReferenceField)
from toxicbuild.core import requests
from toxicbuild.core.plugins import Plugin, PluginMeta
from toxicbuild.core.utils import datetime2string, LoggerMixin
from toxicbuild.master.mail import MailSender
from toxicbuild.master.signals import build_started, build_finished


class PrettyField:
    """A field with a descriptive name for humans"""

    def __init__(self, *args, **kwargs):
        keys = ['pretty_name', 'description']
        for k in keys:
            setattr(self, k, kwargs.get(k))
            try:
                del kwargs[k]
            except KeyError:
                pass

        super().__init__(*args, **kwargs)


class PrettyStringField(PrettyField, StringField):
    pass


class PrettyURLField(PrettyField, URLField):
    pass


class PrettyListField(PrettyField, ListField):
    pass


_translate_table = {PrettyListField: 'list',
                    ListField: 'list',
                    PrettyStringField: 'string',
                    PrettyURLField: 'url',
                    StringField: 'string',
                    UUIDField: 'string',
                    ReferenceField: 'string'}


class MetaMasterPlugin(PluginMeta, DocumentMetaclass):
    """Metaclass that sets name and type to the class definition as
    mongo fields while keeping the interface of setting your plugin's
    name and type as string in definition time."""

    def __new__(cls, name, bases, attrs):
        attrs['_name'] = PrettyStringField(required=True,
                                           default=attrs['name'])
        attrs['_type'] = PrettyStringField(
            required=True, default=attrs['type'])
        attrs['_pretty_name'] = PrettyStringField(
            default=attrs.get('pretty_name'))
        attrs['_description'] = PrettyStringField(default=attrs.get(
            'description'))

        new_cls = super().__new__(cls, name, bases, attrs)
        return new_cls


class MasterPlugin(LoggerMixin, Plugin, EmbeddedDocument,
                   metaclass=MetaMasterPlugin):
    """Base plugin for master's plugins. Master's plugins usually
    react to signals sent by the master."""

    # you must define a name and a type in your own plugin.
    name = 'BaseMasterPlugin'
    pretty_name = ''
    description = "Base for master's plugins"
    type = None

    uuid = UUIDField(default=uuid4)

    meta = {'allow_inheritance': True}

    @classmethod
    def _create_field_dict(cls, field):
        try:
            fdict = {'pretty_name': field.pretty_name,
                     'name': field.name,
                     'type': _translate_table[type(field)]}
        except (KeyError, AttributeError):
            fdict = {'pretty_name': '',
                     'name': field.name,
                     'type': _translate_table[type(field)]}

        return fdict

    @classmethod
    def _translate_schema(cls, fields):
        """Converts the db fields into strings that can be
        serialized."""

        good = copy.copy(fields)
        del good['name']
        del good['type']
        del good['pretty_name']
        del good['description']
        translation = OrderedDict()

        for k, v in good.items():
            translation[k] = cls._create_field_dict(v)

        translation['name'] = fields['name']
        translation['type'] = fields['type']
        translation['pretty_name'] = fields['pretty_name']
        translation['description'] = fields['description']
        return translation

    @classmethod
    def get_schema(cls, to_serialize=False):
        """Returns a dictionary with the schema of the plugin."""

        ordered = cls._fields_ordered
        fields = OrderedDict()

        for of in ordered:
            fields[of] = cls._fields[of]

        fields['type'] = cls.type
        fields['name'] = cls.name
        fields['pretty_name'] = cls.pretty_name
        fields['description'] = cls.description
        del fields['_type']
        del fields['_name']
        del fields['_pretty_name']
        del fields['_description']
        if to_serialize:
            fields = cls._translate_schema(fields)
        return fields

    def to_dict(self):
        schema = type(self).get_schema()
        objdict = {k: getattr(self, k) for k in schema.keys()}
        objdict['uuid'] = str(objdict['uuid'])

        objdict['repository'] = str(self._instance.id)
        return objdict

    async def run(self, sender):
        """Runs the plugin. You must implement this in your plugin."""

        msg = 'You must implement a run() method in your plugin'
        raise NotImplementedError(msg)

    async def stop(self):
        """Stops the plugin. Here is where you may disconnect from signals
        or other stuff needed to stop your plugin."""


class NotificationPlugin(MasterPlugin):

    name = 'NotificationPlugin'
    pretty_name = ''
    description = 'Base plugin for notifications'
    type = 'notification'

    branches = PrettyListField(StringField(), pretty_name="Branches")
    # statuses that trigger the plugin
    statuses = PrettyListField(StringField(), pretty_name="Statuses")

    meta = {'allow_inheritance': True}
    no_list = True
    sender = None

    async def run(self, sender):
        self.sender = sender
        msg = 'running {} for'.format(self.name)
        self.log(msg, level='info')

        if 'running' in self.statuses:
            build_started.connect(self._build_started)

        build_finished.connect(self._build_finished)

    async def stop(self):
        build_started.disconnect(self._build_started)
        build_finished.disconnect(self._build_finished)

    def _build_started(self, repo_id, build):
        ensure_future(self._check_build('started', repo_id, build))

    def _build_finished(self, repo_id, build):
        ensure_future(self._check_build('finished', repo_id, build))

    async def _check_build(self, sig_type, repo_id, build):
        sigs = {'started': self.send_started_message,
                'finished': self.send_finished_message}

        buildset = await build.get_buildset()
        if repo_id == str(self.sender.id) and buildset.branch in self.branches:
            coro = sigs[sig_type]
            ensure_future(coro(self.sender, build))

    async def send_started_message(self, repo, build):
        """Sends a message about a started build. You must implement
        this in your plugin.

        :param repo: A
          :class:`~toxicbuild.master.repository.Repository` instance.
        :param build: A :class:`~toxicbuild.master.build.Build` instance."""

        raise NotImplementedError

    async def send_finished_message(self, repo, build):
        """Sends a message about a finished build. You must implement
        this in your plugin.

        :param repo: A
          :class:`~toxicbuild.master.repository.Repository` instance.
        :param build: A :class:`~toxicbuild.master.build.Build` instance."""

        raise NotImplementedError

    @classmethod
    def _translate_schema(cls, to_serialize=False):
        translation = super()._translate_schema(to_serialize)
        # we move these guys here so the user defined attributes
        # appear first
        translation.move_to_end('branches')
        translation.move_to_end('statuses')
        return translation


class SlackPlugin(NotificationPlugin):
    """Plugin that send notifications about builds to slack."""

    name = 'slack-notification'
    pretty_name = "Slack"
    description = "Sends messages to a slack channel"
    type = 'notification'

    webhook_url = PrettyURLField(required=True, pretty_name='Webhook URL')
    channel_name = PrettyStringField(pretty_name="Channel name")

    def _get_message(self, text):
        return {'text': text, 'channel': self.channel_name,
                'username': 'ToxicBuild'}

    async def _send_message(self, message):
        """Sends a message as an incomming webhook to slack."""

        msg = 'sending message to slack for'
        self.log(msg, level='info')
        headers = {'Content-Type': 'application/json'}
        response = await requests.post(self.webhook_url,
                                       data=json.dumps(message),
                                       headers=headers)
        msg = 'slack response - status {} | text {}'.format(
            response.status, response.text)
        self.log(msg, level='debug')

    async def send_started_message(self, repo, build):

        dt = datetime2string(build.started)
        build_state = 'Build *started* at *{}*'.format(dt)
        title = '[{}] {}'.format(repo.name, build_state)
        msg = self._get_message(title)
        self._send_message(msg)

    async def send_finished_message(self, repo, build):

        if build.status not in self.statuses:
            return

        dt = datetime2string(build.finished)
        build_state = 'Build *finished* at *{}* with status *{}*'.format(
            dt, build.status)
        title = '[{}] {}'.format(repo.name, build_state)

        msg = self._get_message(title)
        await self._send_message(msg)


class EmailPlugin(NotificationPlugin):
    """Sends notification about builds through email"""

    name = 'email-notification'
    pretty_name = 'Email'
    description = 'Sends email messages'
    type = 'notification'

    recipients = PrettyListField(StringField(), pretty_name="Recipients")

    async def send_started_message(self, repo, build):
        buildset = await build.get_buildset()
        dt = datetime2string(build.started)
        subject = '[ToxicBuild][{}] Build started at {}'.format(repo.name, dt)
        message = 'A build has just started for the repository {}.'.format(
            repo.name)

        message += '\n\ncommit: {}\ntitle: {}'.format(buildset.commit,
                                                      buildset.title)

        async with MailSender(self.recipients) as sender:
            await sender.send(subject, message)

    async def send_finished_message(self, repo, build):
        buildset = await build.get_buildset()
        dt = datetime2string(build.finished)
        subject = '[ToxicBuild][{}] Build finished at {}'.format(repo.name, dt)
        message = 'A build finished for the repository {}'.format(repo.name)
        message += '\n\ncommit: {}\ntitle: {}'.format(buildset.commit,
                                                      buildset.title)
        message += '\ntotal time: {}\nstatus: {}'.format(build.total_time,
                                                         build.status)

        async with MailSender(self.recipients) as sender:
            await sender.send(subject, message)


class CustomWebhookPlugin(NotificationPlugin):
    """Sends a POST request to a custom URL. The request
    mime type is json/application and the body of the request
    has a json with 3 keys: ``repository``, ``build`` and ``buildset``."""

    name = 'custom-webhook'
    pretty_name = 'Custom Webhook'
    description = 'Sends messages to a custom webhook.'
    type = 'notification'
    webhook_url = PrettyURLField(required=True, pretty_name='Webhook URL')

    async def _send_message(self, repo, build):
        buildset = await build.get_buildset()
        build_dict = build.to_dict(id_as_str=True)
        buildset_dict = buildset.to_dict(id_as_str=True)
        repo_dict = await repo.to_dict(id_as_str=True)
        msg = {'repo': repo_dict, 'buildset': buildset_dict,
               'build': build_dict}
        await requests.post(self.webhook_url, data=json.dumps(msg))

    async def send_started_message(self, repo, build):
        await self._send_message(repo, build)

    async def send_finished_message(self, repo, build):
        await self._send_message(repo, build)



# -*- coding: utf-8 -*-

# Copyright 2015-2017 Juca Crispim <juca@poraodojuca.net>

# This file is part of toxicbuild.

# toxicbuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# toxicbuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with toxicbuild. If not, see <http://www.gnu.org/licenses/>.


import asyncio
from asyncio import ensure_future
from concurrent.futures import ThreadPoolExecutor
import copy
from datetime import datetime, timezone, timedelta
import fnmatch
import importlib
import logging
import os
import subprocess
import sys
import time
import bcrypt
from toxicbuild.core.exceptions import ExecCmdError, ConfigError


DTFORMAT = '%a %b %d %H:%M:%S %Y %z'


_THREAD_EXECUTOR = ThreadPoolExecutor()


def _get_envvars(envvars):
    """Returns environment variables to be used in shell. Does the
    interpolation of values using the current values from the envvar
    and the values passed as parameters. """

    newvars = copy.copy(os.environ)
    for var, value in envvars.items():
        if var in value:
            current = os.environ.get(var, '')
            value = value.replace(var, current)

        newvars[var] = value

    return newvars


@asyncio.coroutine
def _create_cmd_proc(cmd, cwd, **envvars):
    """Creates a process that will execute a command in a shell.

    :param cmd: command to run.
    :param cwd: Directory to execute the command.
    :param envvars: Environment variables to be used in the command.
    """
    envvars = _get_envvars(envvars)

    proc = yield from asyncio.create_subprocess_shell(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, cwd=cwd,
        env=envvars, preexec_fn=os.setsid)

    return proc


def _kill_group(process):
    """Kills all processes of the group which a process belong.

    :param process: A process that belongs to the group you want to kill.
    """
    try:
        pgid = os.getpgid(process.pid)
        os.killpg(pgid, 9)
    except ProcessLookupError:
        pass


@asyncio.coroutine
def exec_cmd(cmd, cwd, timeout=3600, out_fn=None, **envvars):
    """ Executes a shell command. Raises with the command output
    if return code > 0.
    :param cmd: command to run.
    :param cwd: Directory to execute the command.
    :param timeout: How long we should wait some output. Default
    is 3600.
    :param out_fn: A coroutine that receives each line of the step
    output. The coroutine signature must be in the form:
    mycoro(line_index, line).
    :param envvars: Environment variables to be used in the command.
    """

    proc = yield from _create_cmd_proc(cmd, cwd, **envvars)
    out = []

    line_index = 0
    while proc.returncode is None or not out:
        outline = yield from asyncio.wait_for(proc.stdout.readline(), timeout)
        outline = outline.decode()
        if out_fn:
            ensure_future(out_fn(line_index, outline))

        line_index += 1
        out.append(outline)

    output = ''.join(out).strip('\n')
    # we must ensure that all process started by our command are
    # dead.
    _kill_group(proc)
    if int(proc.returncode) > 0:
        raise ExecCmdError(output)

    return output


def load_module_from_file(filename):
    """ Load a module from a source file
    :param filename: full path for file to be loaded.
    """
    fname, extension = filename.rsplit('.', 1)
    fname = fname.rsplit(os.sep, 1)[-1]
    source_file = importlib.machinery.SourceFileLoader(fname, filename)

    try:
        module = source_file.load_module()
    except FileNotFoundError:
        err_msg = 'Config file "%s" does not exist!' % (filename)
        raise FileNotFoundError(err_msg)
    except Exception as e:
        err_msg = 'There is something wrong with your file. '
        err_msg += 'The original exception was:\n{}'.format(e.args[0])
        raise ConfigError(err_msg)

    return module


def log(msg, level='info'):
    log = getattr(logging, level)
    dt = now().strftime('%Y-%m-%d %H:%M:%S')
    msg = ' {} - {}'.format(dt, msg)
    log(msg)


class LoggerMixin:

    """A simple mixin to use log on a class."""

    def log(self, msg, level='info'):
        """Appends the class name before the log message. """

        log('[{}] {} '.format(type(self).__name__, msg), level)


def format_timedelta(td):
    """Format a timedelta object to a human-friendly string.

    :param dt: A timedelta object."""

    return str(td).split('.')[0]


def datetime2string(dt, dtformat=DTFORMAT):
    """Transforms a datetime object into a formated string.

    :param dt: The datetime object.
    :param dtformat: The format to use."""

    if dt.utcoffset() is None:
        tz = timezone(timedelta(seconds=0))
        dt = dt.replace(tzinfo=tz)
    return datetime.strftime(dt, dtformat)


def string2datetime(dtstr, dtformat=DTFORMAT):
    """Transforns a string into a datetime object acording to ``dtformat``.

    :param dtstr: The string containing the formated date.
    :param dtformat: The format of the formated date.
    """
    return datetime.strptime(dtstr, dtformat)


def utc2localtime(utcdatetime):
    """Transforms a utc datetime object into a datetime object
    in local time.

    :param utcdatetime: A datetime object"""

    off = time.localtime().tm_gmtoff
    td = timedelta(seconds=off)
    tz = timezone(td)
    local = utcdatetime + td
    localtime = datetime(local.year, local.month, local.day,
                         local.hour, local.minute, local.second,
                         local.microsecond,
                         tzinfo=tz)
    return localtime


def localtime2utc(localdatetime):
    """Transforms a local datetime object into a datetime object
    in utc time.

    :param localdatetime: A datetime object."""
    off = time.localtime().tm_gmtoff
    td = timedelta(seconds=off)
    utc = localdatetime - td
    utctz = timezone(timedelta(seconds=0))
    utctime = datetime(utc.year, utc.month, utc.day,
                       utc.hour, utc.minute, utc.second,
                       utc.microsecond,
                       tzinfo=utctz)
    return utctime


def now():
    """ Returns the localtime with timezone info. """

    off = time.localtime().tm_gmtoff
    tz = timezone(timedelta(seconds=off))
    return datetime.now(tz=tz)


def get_toxicbuildconf(directory):
    """Returns the toxicbuild.conf module.

    :param directory: Directory to look for toxicbuild.conf"""

    configfile = os.path.join(directory, 'toxicbuild.conf')
    return load_module_from_file(configfile)


def list_builders_from_config(confmodule, branch, slave):
    builders = [b['name'] for b in confmodule.BUILDERS
                if (b.get('branch') == branch or b.get('branch')is None) and
                (b.get('slave') == slave.name or b.get('slave') is None)]
    return builders


# inherit_docs thanks to Raymond Hettinger on stackoverflow
# stackoverflow.com/questions/8100166/inheriting-methods-docstrings-in-python
def inherit_docs(cls):
    """ Inherit docstrings from base classes' methods.
    Can be used as a decorator

    :param cls: Class that will inherit docstrings from its parents.
    """
    for name, func in vars(cls).items():
        if not func.__doc__:
            for parent in cls.__bases__:
                try:
                    parfunc = getattr(parent, name)
                except Exception:
                    continue
                if parfunc and getattr(parfunc, '__doc__', None):
                    func.__doc__ = parfunc.__doc__
                    break
    return cls


@asyncio.coroutine
def read_stream(reader):
    """ Reads the input stream. First reads the bytes until the first "\\n".
    These first bytes are the length of the full message.

    :param reader: An instance of :class:`asyncio.StreamReader`
    """

    data = yield from reader.read(1)
    if not data or data == b'\n':
        raw_data = b''
    else:
        char = None
        while char != b'\n' and char != b'':
            char = yield from reader.read(1)
            data += char

        len_data = int(data)
        if len_data <= 1024:
            raw_data = yield from reader.read(len_data)
        else:
            raw_data = yield from reader.read(1024)

        while len(raw_data) < len_data:
            left = len_data - len(raw_data)
            next_chunk = left if left < 1024 else 1024
            raw_data += yield from reader.read(next_chunk)

    return raw_data


@asyncio.coroutine
def write_stream(writer, data):
    """ Writes ``data`` to output. Encodes data to utf-8 and prepend the
    lenth of the data before sending it.

    :param writer: An instance of asyncio.StreamWriter
    :param data: String data to be sent.
    """

    data = data.encode('utf-8')
    data = '{}\n'.format(len(data)).encode('utf-8') + data
    writer.write(data)
    yield from writer.drain()


def bcrypt_string(string, salt):
    encoding = sys.getdefaultencoding()
    if isinstance(salt, str):
        salt = salt.encode(encoding)
    encrypted = bcrypt.hashpw(string.encode(encoding), salt)
    return encrypted.decode()


class changedir(object):

    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.current_dir = path

    def __enter__(self):
        os.chdir(self.current_dir)

    def __exit__(self, *a, **kw):
        os.chdir(self.old_dir)


def match_string(smatch, filters):
    """Checks if a string match agains a list
    of filters containing wildcards.

    :param smatch: String to test against the filters
    :param filters: Filter to match a string."""

    return any([fnmatch.fnmatch(smatch, f) for f in filters])


class MatchKeysDict(dict):
    """A dictionary that returns the values matching the keys using
    :meth:`toxicbuild.core.utils.match_string`.

    .. code-block:: python

        >>> d = MatchKeysDict()
        >>> d['k*'] = 1
        >>> d['key']
        1
        >>> k['keyboard']
        1
    """

    def __getitem__(self, key):
        for k in self.keys():
            if match_string(key, [k]):
                return super().__getitem__(k)
        return super().__getitem__(key)


@asyncio.coroutine
def run_in_thread(fn, *args, **kwargs):
    return _THREAD_EXECUTOR.submit(fn, *args, **kwargs)

# Sorry, but not willing to test  a daemonizer.


def daemonize(call, cargs, ckwargs, stdout, stderr,
              workdir, pidfile):  # pragma: no cover
    """ Run a callable as a daemon

    :param call: a callable.
    :param cargs: args to ``call``.
    :param ckwargs: kwargs to ``call``.
    :param stdout: daemon's stdout.
    :param stderr: daemon's stderr.
    :param workdir: daemon's workdir
    :param pidfile: pidfile's path
    """
    _create_daemon(stdout, stderr, workdir)
    pid = os.getpid()
    with open(pidfile, 'w') as f:
        f.write(str(pid))

    call(*cargs, **ckwargs)


def _create_daemon(stdout, stderr, workdir):  # pragma: no cover
    _fork_off_and_die()
    os.setsid()
    _fork_off_and_die()
    os.umask(0)
    os.chdir(workdir)
    _redirect_file_descriptors(stdout, stderr)


def _fork_off_and_die():  # pragma: no cover
    pid = os.fork()
    if pid != 0:
        sys.exit(0)


def _redirect_file_descriptors(stdout, stderr):  # pragma: no cover
    for fd in sys.stdout, sys.stderr:
        fd.flush()

    sys.stdout = open(stdout, 'a', 1)
    sys.stderr = open(stderr, 'a', 1)
